import matplotlib.cm as cm
import matplotlib.pylab as plt
import numpy as np
from xarray import DataArray

from indica.available_quantities import READER_QUANTITIES
from indica.converters import TransectCoordinates
from indica.models.abstract_diagnostic import AbstractDiagnostic
from indica.numpy_typing import LabeledArray
from indica.utilities import build_dataarrays


class ThomsonScattering(AbstractDiagnostic):
    """
    Object representing a Thomson scattering diagnostic
    """

    def __init__(
        self,
        name: str,
        instrument_method="get_thomson_scattering",
    ):
        self.transform: TransectCoordinates
        self.name = name
        self.instrument_method = instrument_method
        self.quantities = READER_QUANTITIES[self.instrument_method]

    def _build_bckc_dictionary(self):
        bckc = {
            "t": self.t,
            "channel": np.arange(len(self.transform.x1)),
            "ne": self.Ne_at_channels,
            "te": self.Te_at_channels,
            "x": self.transform.x,
            "y": self.transform.y,
            "z": self.transform.z,
            "R": self.transform.R,
        }
        self.bckc = build_dataarrays(bckc, self.quantities, transform=self.transform)

    def __call__(
        self,
        Ne: DataArray = None,
        Te: DataArray = None,
        t: LabeledArray = None,
        calc_rho: bool = False,
        **kwargs,
    ):
        """
        Calculate diagnostic measured values

        Parameters
        ----------
        Ne
            Electron density profile (dims = "rho", "t")
        Te
            Electron temperature profile (dims = "rho", "t")

        Returns
        -------
        Dictionary of back-calculated quantities (identical to datareader.py)

        """
        if self.plasma is not None:
            if t is None:
                t = self.plasma.time_to_calculate
            Ne = self.plasma.electron_density.interp(t=t)
            Te = self.plasma.electron_temperature.interp(t=t)
        else:
            if Ne is None or Te is None:
                raise ValueError("Give inputs of assign plasma class!")

        self.t = t
        self.Ne = Ne
        self.Te = Te

        Ne_at_channels = self.transform.map_profile_to_rho(
            Ne,
            t=self.t,
            calc_rho=calc_rho,
        )
        Ne_at_channels = Ne_at_channels.assign_coords(
            R=("channel", self.transform.R.data)
        )
        Ne_at_channels = Ne_at_channels.assign_coords(
            z=("channel", self.transform.z.data)
        )
        Te_at_channels = self.transform.map_profile_to_rho(
            Te,
            t=self.t,
            calc_rho=calc_rho,
        )
        Te_at_channels = Te_at_channels.assign_coords(
            R=("channel", self.transform.R.data)
        )
        Te_at_channels = Te_at_channels.assign_coords(
            z=("channel", self.transform.z.data)
        )

        self.Ne_at_channels = Ne_at_channels
        self.Te_at_channels = Te_at_channels

        self._build_bckc_dictionary()

        return self.bckc

    def plot(self):
        if len(self.bckc) == 0:
            print("No model results to plot")
            return

        # Back-calculated profiles
        cols_time = cm.gnuplot2(np.linspace(0.1, 0.75, np.size(self.t), dtype=float))
        plt.figure()
        for i, t in enumerate(self.t):
            Ne = self.bckc["ne"].sel(t=t, method="nearest")
            rho = Ne.transform.rhop.sel(t=t, method="nearest")
            plt.scatter(
                rho,
                Ne,
                color=cols_time[i],
                marker="o",
                alpha=0.7,
                label=f"t={t:1.2f} s",
            )
        plt.xlabel("Channel")
        plt.ylabel("Measured electron density (m^-3)")
        plt.legend()

        plt.figure()
        for i, t in enumerate(self.t):
            Te = self.bckc["te"].sel(t=t, method="nearest")
            rho = Te.transform.rhop.sel(t=t, method="nearest")
            plt.scatter(
                rho,
                Te,
                color=cols_time[i],
                marker="o",
                alpha=0.7,
                label=f"t={t:1.2f} s",
            )
        plt.xlabel("Channel")
        plt.ylabel("Measured electron temperature (eV)")
        plt.legend()



"""Base class for IO objects.
"""

from abc import ABC
from abc import abstractmethod
from typing import Literal


class BaseIO(ABC):
    """An abstract class defining methods needed by all IO objects."""

    def __enter__(self) -> "BaseIO":
        """Called at beginning of a context manager."""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> Literal[False]:
        """Close reader at end of context manager. Don't try to handle
        exceptions."""
        self.close()
        return False

    def authenticate(self, name: str, password: str) -> bool:
        """Confirms user has permission to write data.

        This must be called before writing data to some locations. The
        default implementation does nothing. If the value of
        `py:meth:requires_authentication` is ``False`` then it does
        not need to be called.

        Parameters
        ----------
        name
            Username to authenticate against.
        password
            Password for that user.

        Returns
        -------
        :
            Indicates whether authentication was succesful.

        """
        return True

    @property
    @abstractmethod
    def requires_authentication(self) -> bool:
        """Indicates whether authentication is required to read data.

        Returns
        -------
        :
            True if authentication is needed, otherwise false.
        """
        raise NotImplementedError(
            "{} does not implement a "
            "'requires_authentication' "
            "property.".format(self.__class__.__name__)
        )

    @abstractmethod
    def close(self) -> None:
        """Closes connection to whatever backend (file, database, server,
        etc.) to which data is being written."""
        raise NotImplementedError(
            "{} does not implement a 'close' method.".format(self.__class__.__name__)
        )



"""Various miscellanious helper functions."""
from copy import deepcopy
from getpass import getuser
import hashlib
import inspect
import os
import string
from typing import Any
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import List
from typing import Optional
from typing import Tuple

from matplotlib import cm
from matplotlib import rcParams
import matplotlib.pylab as plt
import numpy as np
import periodictable
from scipy.interpolate import CubicSpline
import xarray as xr
from xarray import apply_ufunc
from xarray import DataArray

from indica.datatypes import DATATYPES
from indica.datatypes import UNITS
from .numpy_typing import ArrayLike
from .numpy_typing import LabeledArray
from .numpy_typing import OnlyArray

DATA_PATH = f"/home/{getuser()}/data/Indica/"
FIG_PATH = f"/home/{getuser()}/figures/Indica/"
CACHE_DIR = ".indica"


def get_element_info(element: str) -> Tuple[int, float, str, str]:
    """
    Return periodic table information of specified element

    Parameters
    ----------
    element
        Element name (e.g. molybdenum) or symbol (Mo)

    Returns
    -------
    atomic number, atomic weight, element name
    """
    number: int
    mass: float
    name: str
    symbol: str
    if len(element) <= 2:
        _element = element.capitalize()
    else:
        _element = element.lower()

    if hasattr(periodictable, _element):
        elem_info = getattr(periodictable, _element)
        number, mass, name, symbol = (
            elem_info.number,
            elem_info.mass,
            elem_info.name,
            elem_info.symbol,
        )
    else:
        number, mass, name, symbol = 0, 0.0, "", ""

    return number, mass, name, symbol


def positional_parameters(func: Callable[..., Any]) -> Tuple[List[str], Optional[str]]:
    """Returns an ordered list of the names of arguments which can be
    passed to a function positionally.

    Parameters
    ----------
    func
        A function/callable for which to get information on the positional
        arguments.

    Returns
    -------
    param_names : List[str]
        Ordered list of names of function parameters which can be passed to
        positionally.
    var_positional : str or None
        Name of variable positional parameter, if present (e.g., ``*args``).

    """
    param_names = []
    var_positional = None
    for param in inspect.signature(func).parameters.values():
        if (
            param.kind == inspect.Parameter.POSITIONAL_ONLY
            or param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD
        ):
            param_names.append(param.name)
        elif param.kind == inspect.Parameter.VAR_POSITIONAL:
            var_positional = param.name
        else:
            break
    return param_names, var_positional


def sum_squares(x: OnlyArray, axis: int, **kwargs: Any) -> OnlyArray:
    """Perform a reduction on the provided data by taking the sum of squares.

    Parameters
    ----------
    x
        The arrayto be reduced.
    axis
        The axis along which to perform the reduction.
    **kwargs
        Additiona keyword arguments (unused)

    """
    return np.sum(x**2, axis=axis)


def get_slice_limits(low: float, high: float, data: OnlyArray) -> Tuple[int, int]:
    """Returns the start and end points needed to slice ``data`` so that
    all values fall between ``high`` and ``low`` (inclusive).

    An error will be raised if ``data`` does not contain values above and below
    these limits.

    Parameters
    ----------
    low
        The lower limit for the data.
    high
        The upper limit for the data.
    data
        An ordered 1-D array of values.

    Returns
    -------
    start
        The index above which all values of ``data`` are greater or equal to
        ``low``.
    end
        The index below which all values of ``data`` are less than or equal to
        ``high"

    """
    start = np.argmax(data > low) - 1
    if start < 0:
        raise ValueError("Low value {} not in range of provided " "data.".format(low))
    end = np.argmax(data >= high)
    if end < 1:
        raise ValueError("High value {} not in range of provided " "data.".format(high))

    return (start, end)  # type: ignore


def to_filename(name: str) -> str:
    """Takes a string and returns a valid filename based on it."""
    valid_chars = "-_.(){}{}".format(string.ascii_letters, string.digits)
    filename = name.replace("/", "-").replace("\\", "-")
    filename = filename.replace(" ", "_")
    filename = "".join(c for c in filename if c in valid_chars)
    return filename


def coord_array(coord_vals: ArrayLike, coord_name: str):
    """Create a DataArray which can be used for a coordinate system.

    Parameters
    ----------
    coord_vals
        The position/value at each point on the coordinate grid.
    coord_name
        The name of the dimension.
    """
    datatype = (
        "major_rad"
        if coord_name == "R"
        else "time"
        if coord_name == "t"
        else "norm_flux_pol"
        if coord_name == "rhop"
        else "norm_flux_tor"
        if coord_name == "rhot"
        else coord_name,
        "plasma",
    )
    return DataArray(
        coord_vals, coords=[(coord_name, coord_vals)], attrs={"datatype": datatype}
    )


def broadcast_spline(
    spline: CubicSpline,
    spline_dims: Tuple,
    spline_coords: Dict[Hashable, Any],
    interp_coord: DataArray,
):
    """Applies to a `:class:xarray.DataArray` input, broadcasting and/or
    interpolating appropriately.

    Note
    ----
    Currently the only dimension which are checked for presence in
    both the spline data and the input is "t".

    TODO: Implement these checks for other dimensions as well.

    """
    if "t" in interp_coord.coords and "t" in spline_dims:
        time_outer_product = apply_ufunc(
            spline,
            interp_coord,
            input_core_dims=[[]],
            output_core_dims=[tuple(d if d != "t" else "__old_t" for d in spline_dims)],
        ).assign_coords(__old_t=coord_array(spline_coords["t"].data, "__old_t"))
        result = time_outer_product.indica.interp2d(
            __old_t=interp_coord.coords["t"],
            method="cubic"
            if time_outer_product.coords["__old_t"].size > 3
            else "linear",
        ).assign_coords({k: v for k, v in spline_coords.items() if k != "t"})
        del result.coords["__old_t"]
        return result
    else:
        return apply_ufunc(
            spline,
            interp_coord,
            input_core_dims=[[]],
            output_core_dims=[spline_dims],
        ).assign_coords({k: v for k, v in spline_coords.items()})


def build_dataarrays(
    data: Dict[str, LabeledArray],
    available_quantities: Dict[str, Tuple[str, list]],
    tstart: float = None,
    tend: float = None,
    transform=None,
    include_error: bool = True,
    verbose: bool = False,
) -> Dict[str, DataArray]:
    """Organizes data in DataArray format with coordinates, long_name & units"""
    data_arrays: dict = {}
    for quantity in available_quantities.keys():
        if quantity not in data.keys():
            continue

        # Build coordinate dictionary
        datatype, dims = available_quantities[quantity]
        if verbose:
            print(f"  {quantity} - {datatype}")

        coords: dict = {}
        for dim in dims:
            coords[dim] = data[dim]

        # Build DataArray
        _data = format_dataarray(data[quantity], datatype, coords)
        if "t" in _data.dims and tstart is not None and tend is not None:
            _data = _data.sel(t=slice(tstart, tend))
            _data = _data.sortby([dim for dim in dims if dim != "t"])

        # Build error DataArray and assign as coordinate
        if include_error and len(dims) != 0:
            _error = xr.zeros_like(_data)
            if quantity + "_error" in data:
                _error = format_dataarray(data[quantity + "_error"], datatype, coords)
                if "t" in _error.dims and tstart is not None and tend is not None:
                    _error = _error.sel(t=slice(tstart, tend))
            _data = _data.assign_coords(error=(_data.dims, _error.data))

        # Check that times are unique
        if "t" in _data:
            t_unique, ind_unique = np.unique(_data["t"], return_index=True)
            if len(_data["t"]) != len(t_unique):
                _data = _data.isel(t=ind_unique)

        # Add attributes and assign to dictionary
        if transform is not None:
            _data.attrs["transform"] = transform
        if "uid" in _data:
            _data.attrs["uid"] = data["uid"]
        if "revision" in _data:
            _data.attrs["revision"] = data["revision"]

        data_arrays[quantity] = _data
    return data_arrays


def format_coord(data: LabeledArray, datatype: str):
    """
    Create coordinate dataarray using the variable name == dimension
    and the values as coordinates

    Parameters
    ----------
    data
        Input coordinate data to convert to DataArray
    var_name
        Name of the variable to be assigned as coordinate
    """
    return format_dataarray(data, datatype, {datatype: data})


def format_dataarray(
    data: LabeledArray,
    datatype: str,
    coords: Dict[str, Any] = {},
    make_copy: bool = False,
):
    """
    Generate DataArray with long_name & units (indica/datatypes.py) & coordinates

    Parameters
    ----------
    data
        Input data
    datatype
        Variable data type name (see DATATYPES)
    coords
        Coordinates sequence (see xr.DataArray documentation)

    Returns
    -------
    Formatted data array, including attribute assignement (also to coords)

    """

    if make_copy:
        _data = deepcopy(data)
    else:
        _data = data

    if len(coords) != 0:
        processed_coords = {
            name: coord.data if isinstance(coord, DataArray) else coord
            for name, coord in coords.items()
        }
        data_array = DataArray(_data, coords=processed_coords, name=datatype)
    else:
        data_array = DataArray(_data, name=datatype)

    assign_datatype(data_array, datatype)
    for dim in data_array.dims:
        assign_datatype(data_array.coords[dim], dim)

    return data_array


def assign_datatype(
    data_array: DataArray,
    datatype: str,
):
    try:
        long_name, units_key = DATATYPES[datatype]
        units = UNITS[units_key]
    except Exception:
        raise Exception(f"\n Check DATATYPE and UNITS for {datatype}. \n")

    data_array.attrs["long_name"] = long_name
    data_array.attrs["units"] = units


def print_like(string: str):
    print(f"\n {string}")


def get_function_name():
    return str(inspect.stack()[1][3])


def _rect_inter_inner(x1, x2):
    n1 = x1.shape[0] - 1
    n2 = x2.shape[0] - 1
    X1 = np.c_[x1[:-1], x1[1:]]
    X2 = np.c_[x2[:-1], x2[1:]]
    S1 = np.tile(X1.min(axis=1), (n2, 1)).T
    S2 = np.tile(X2.max(axis=1), (n1, 1))
    S3 = np.tile(X1.max(axis=1), (n2, 1)).T
    S4 = np.tile(X2.min(axis=1), (n1, 1))
    return S1, S2, S3, S4


def _rectangle_intersection_(x1, y1, x2, y2):
    S1, S2, S3, S4 = _rect_inter_inner(x1, x2)
    S5, S6, S7, S8 = _rect_inter_inner(y1, y2)

    C1 = np.less_equal(S1, S2)
    C2 = np.greater_equal(S3, S4)
    C3 = np.less_equal(S5, S6)
    C4 = np.greater_equal(S7, S8)

    ii, jj = np.nonzero(C1 & C2 & C3 & C4)
    return ii, jj


def intersection(x1, y1, x2, y2):
    """
    INTERSECTIONS Intersections of curves.
       Computes the (x,y) locations where two curves intersect.  The curves
       can be broken with NaNs or have vertical segments.
    usage:
    x,y=intersection(x1,y1,x2,y2)
        Example:
        a, b = 1, 2
        phi = np.linspace(3, 10, 100)
        x1 = a*phi - b*np.sin(phi)
        y1 = a - b*np.cos(phi)
        x2=phi
        y2=np.sin(phi)+2
        x,y, ix, iy=intersection(x1,y1,x2,y2)
        plt.plot(x1,y1,c='r')
        plt.plot(x2,y2,c='g')
        plt.plot(x,y,'*k')
        plt.show()
    """
    ii, jj = _rectangle_intersection_(x1, y1, x2, y2)
    n = len(ii)

    dxy1 = np.diff(np.c_[x1, y1], axis=0)
    dxy2 = np.diff(np.c_[x2, y2], axis=0)

    T = np.zeros((4, n))
    AA = np.zeros((4, 4, n))
    AA[0:2, 2, :] = -1
    AA[2:4, 3, :] = -1
    AA[0::2, 0, :] = dxy1[ii, :].T
    AA[1::2, 1, :] = dxy2[jj, :].T

    BB = np.zeros((4, n))
    BB[0, :] = -x1[ii].ravel()
    BB[1, :] = -x2[jj].ravel()
    BB[2, :] = -y1[ii].ravel()
    BB[3, :] = -y2[jj].ravel()

    for i in range(n):
        try:
            T[:, i] = np.linalg.solve(AA[:, :, i], BB[:, i])
        except np.linalg.LinAlgError:
            T[:, i] = np.NaN

    in_range = (T[0, :] >= 0) & (T[1, :] >= 0) & (T[0, :] <= 1) & (T[1, :] <= 1)

    xy0 = T[2:, in_range]
    xy0 = xy0.T

    indii = ii[in_range] + T[0, in_range]
    indjj = jj[in_range] + T[1, in_range]

    return xy0[:, 0], xy0[:, 1], indii, indjj


def check_time_present(t_desired: LabeledArray, t_array: LabeledArray):
    equil_ok = (np.min(t_desired) >= np.min(t_array)) * (
        np.max(t_desired) <= np.max(t_array)
    )
    if not equil_ok:
        raise ValueError(f"Desired time {t_desired} not available in array {t_array}")


def save_figure(
    path_name: str = "",
    fig_name: str = "",
    orientation: str = "landscape",
    dpi: int = 300,
    quality: int = 95,
    ext: str = "png",
    save_fig: bool = True,
    create_path: bool = True,
):
    if save_fig:
        _fig_name = deepcopy(fig_name)
        _path_name = deepcopy(path_name)
        if _path_name[-1] != "/":
            _path_name = f"{_path_name}/"
        _file = f"{_path_name}{_fig_name}.{ext}"

        kwargs = {"orientation": orientation, "dpi": dpi}
        if ext != "svg":
            kwargs["pil_kwargs"] = {"quality": quality}

        if create_path and not os.path.exists(_path_name):
            os.mkdir(_path_name)
            print(f"Creating directory {_path_name}")

        plt.savefig(
            _file,
            **kwargs,
        )
        print(f"Saving picture to {_file}")


def set_plot_colors(
    color_map: str = "gnuplot2",
):

    cmap = getattr(cm, color_map)
    colors = {
        "electron": cmap(0.1),
        "ion": cmap(0.75),
        "fast": cmap(0.4),
        "impurity": cmap(0.0),
        "raw": "black",
        "thermal": "red",
        "total": "black",
        "binned": cmap(0.2),
        "bckc": cmap(0.75),
    }

    return cmap, colors


def set_plot_rcparams(option: str = "profiles"):
    plot_params: dict = {
        "profiles": {
            "font.size": 13,
            "legend.fontsize": 11,
            "lines.markersize": 10,
            "lines.linewidth": 2,
        },
        "multi": {
            "font.size": 13,
            "legend.fontsize": 13,
            "lines.markersize": 10,
            "lines.linewidth": 2,
            "figure.figsize": [6.4, 3.8],
            "figure.subplot.bottom": 0.15,
        },
        "time_evolution": {
            "font.size": 11,
            "legend.fontsize": 8,
            "lines.markersize": 5,
            "lines.linewidth": 2,
            "font.weight": 600,
        },
    }

    if option not in plot_params.keys():
        return

    for key, value in plot_params[option].items():
        rcParams.update({key: value})


def set_axis_sci(plot_object=None, axis: str = "y"):
    if hasattr(plot_object, "colorbar"):
        plot_object.colorbar.formatter.set_powerlimits((0, 0))
    else:
        if plot_object is None:
            plot_object = plt
        plot_object.ticklabel_format(style="sci", axis=axis, scilimits=(0, 0))


def hash_vals(**kwargs: Any) -> str:
    """Produces an SHA256 hash from the key-value pairs passed as
    arguments.

    Parameters
    ---------
    kwargs
        The data to use for the hash.

    Returns
    -------
    str
        A hexadecimal representation of the hash.
    """
    # TODO: include date/time in hash
    hash_result = hashlib.sha256()
    for key, val in kwargs.items():
        hash_result.update(bytes(key, encoding="utf-8"))
        hash_result.update(b":")
        hash_result.update(bytes(str(val), encoding="utf-8"))
        hash_result.update(b",")
    return hash_result.hexdigest()



from copy import deepcopy

from matplotlib import cm
import matplotlib.pylab as plt
import numpy as np
import xarray as xr
from xarray import DataArray

from indica.available_quantities import READER_QUANTITIES
from indica.converters import LineOfSightTransform
from indica.models.abstract_diagnostic import AbstractDiagnostic
from indica.numpy_typing import LabeledArray
import indica.physics as ph
from indica.utilities import build_dataarrays
from indica.utilities import format_coord
from indica.utilities import set_axis_sci
from indica.utilities import set_plot_rcparams


class BremsstrahlungDiode(AbstractDiagnostic):
    """
    Object representing an diode filter diagnostic measuring
    Bremsstrahlung radiation in a specified wavelength range

    TODO: currently working only for Bremsstrahlung emission!!!
    """

    los_integral: DataArray

    def __init__(
        self,
        name: str,
        filter_wavelength: float = 531.5,
        filter_fwhm: float = 1,
        filter_type: str = "boxcar",
        etendue: float = 1.0,
        calibration: float = 1,
        instrument_method="get_diode_filters",
        channel_mask: slice = None,
    ):
        """
        Filtered diode diagnostic measuring Bremsstrahlung
        TODO: does not account for contaminating spectral lines

        Parameters
        ----------
        name
            Diagnostic name
        filter_wavelength
            Central wavelength of transmission filter
        filter_fwhm
            FWHM of transmission filter
        filter_type
            Function describing filter shape
        etendue
            Etendue of system
        calibration
            Absolute calibration to convert from Watts -> counts
        instrument_method
            Name of indica reader method to read experimental diagnostic data
        """
        self.transform: LineOfSightTransform
        self.name = name
        self.filter_wavelength = filter_wavelength
        self.filter_fwhm = filter_fwhm
        self.filter_type = filter_type
        self.etendue = etendue
        self.calibration = calibration
        self.instrument_method = instrument_method
        self.quantities = READER_QUANTITIES[self.instrument_method]
        self.channel_mask = channel_mask

        wavelength = np.linspace(
            self.filter_wavelength - self.filter_fwhm * 2,
            self.filter_wavelength + self.filter_fwhm * 2,
        )
        self.wavelength = format_coord(wavelength, "wavelength")

        # Transmission filter function
        transmission = ph.make_window(
            wavelength,
            self.filter_wavelength,
            self.filter_fwhm,
            window=self.filter_type,
        )
        self.transmission = DataArray(transmission, coords={"wavelength": wavelength})

    def integrate_spectra(self, spectra: DataArray, fit_background: bool = True):
        """
        Apply the diode transmission function to an input spectra

        Parameters
        ----------
        spectra
            Spectra with dimensions (channel, wavelength, time) in any order,
            and with units of W/m**2
        fit_background
            If True - background fitted and then integrated
            If False - spectra integrated using filter without any fitting

        Returns
        -------
        Background emission & integral of spectra using the filter transmission
        TODO: uncertainty on fit not calculated
        TODO: move spectral fitting to separate method outside of the class
        """

        # Interpolate transmission filter on spectral wavelength & restrict to > 0
        _transmission = self.transmission.interp(wavelength=spectra.wavelength)
        transmission = _transmission.where(_transmission > 1.0e-3, drop=True)
        wavelength_slice = slice(
            np.min(transmission.wavelength), np.max(transmission.wavelength)
        )

        _spectra = spectra.sortby("wavelength").sel(wavelength=wavelength_slice)
        if hasattr(_spectra, "error"):
            _spectra_err = spectra.error.sortby("wavelength").sel(
                wavelength=wavelength_slice
            )
        else:
            _spectra_err = xr.full_like(_spectra, 0.0)

        # Fit spectra to calculate background emission, filter and integrate
        if fit_background:
            fit = _spectra.polyfit("wavelength", 0)
            _spectra_fit = fit.polyfit_coefficients.sel(degree=0)
            spectra_to_integrate = _spectra_fit.expand_dims(
                dim={"wavelength": _spectra.wavelength}
            )
            spectra_to_integrate_err = xr.full_like(spectra_to_integrate, 0.0)
        else:
            spectra_to_integrate = _spectra
            spectra_to_integrate_err = _spectra_err

        spectra_to_integrate = spectra_to_integrate.assign_coords(
            error=(spectra_to_integrate.dims, spectra_to_integrate_err.data)
        )

        integral = (spectra_to_integrate * transmission).sum("wavelength")
        integral_err = (np.sqrt((spectra_to_integrate_err * transmission) ** 2)).sum(
            "wavelength"
        )
        integral = integral.assign_coords(error=(integral.dims, integral_err.data))

        return spectra_to_integrate, integral

    def _build_bckc_dictionary(self):
        bckc = {
            "t": self.t,
            "channel": np.arange(len(self.transform.x1)),
            "location": self.transform.origin,
            "direction": self.transform.direction,
            "brightness": self.los_integral,
        }
        self.bckc = build_dataarrays(bckc, self.quantities, transform=self.transform)

    def __call__(
        self,
        Te: DataArray = None,
        Ne: DataArray = None,
        Zeff: DataArray = None,
        t: LabeledArray = None,
        calc_rho: bool = False,
        **kwargs,
    ):
        """
        Calculate Bremsstrahlung emission and model measurement

        TODO: add set of spectral lines to model different line diodes

        Parameters
        ----------
        Te
            electron temperature
        Ne
            electron density
        Zeff
            Total effective charge
        t
            time
        TODO: emission needs a new name as it's in units [W m**-2 nm**-1]
        """

        if self.plasma is not None:
            if t is None:
                t = self.plasma.time_to_calculate
            Ne = self.plasma.electron_density.sel(t=t)
            Te = self.plasma.electron_temperature.sel(t=t)
            Zeff = self.plasma.zeff.sel(t=t).sum("element")
        else:
            if Ne is None or Te is None or Zeff is None:
                raise ValueError("Give inputs of assign plasma class!")

        self.t: DataArray = t
        self.Te: DataArray = Te
        self.Ne: DataArray = Ne
        self.Zeff: DataArray = Zeff

        # Bremsstrahlung emission for each time, radial position and wavelength
        wlength = deepcopy(self.wavelength)
        for dim in Ne.dims:
            wlength = wlength.expand_dims(dim={dim: self.Ne[dim]})
        self.emission = ph.zeff_bremsstrahlung(Te, Ne, wlength, zeff=Zeff)
        self.emissivity = (self.emission * self.transmission).integrate("wavelength")
        los_integral = self.transform.integrate_on_los(
            self.emissivity,
            t=t,
            calc_rho=calc_rho,
        )
        if self.channel_mask is not None:
            los_integral = los_integral.where(
                (los_integral.channel > self.channel_mask.start)
                & (los_integral.channel < self.channel_mask.stop)
            )

        self.los_integral = los_integral

        self._build_bckc_dictionary()
        return self.bckc

    def plot(self, nplot: int = 1):
        set_plot_rcparams("profiles")
        if len(self.bckc) == 0:
            print("No model results to plot")
            return

        # Line-of-sight information
        self.transform.plot(np.mean(self.t))

        # Back-calculated profiles
        cols_time = cm.gnuplot2(np.linspace(0.1, 0.75, len(self.t), dtype=float))
        plt.figure()
        for i, t in enumerate(np.array(self.t)):
            if i % nplot:
                continue

            _brightness = self.bckc["brightness"].sel(t=t, method="nearest")
            if "beamlet" in _brightness.dims:
                plt.fill_between(
                    _brightness.channel,
                    _brightness.max("beamlet"),
                    _brightness.min("beamlet"),
                    color=cols_time[i],
                    alpha=0.5,
                )
                brightness = _brightness.mean("beamlet")
            else:
                brightness = _brightness
            brightness.plot(label=f"t={t:1.2f} s", color=cols_time[i])
        set_axis_sci()
        plt.title(self.name.upper())
        plt.xlabel("Channel")
        plt.ylabel("Measured brightness (W/m^2)")
        plt.legend()

        # Local emissivity profiles
        plt.figure()
        for i, t in enumerate(np.array(self.t)):
            if i % nplot:
                continue
            plt.plot(
                self.emissivity.rhop,
                self.emissivity.sel(t=t),
                color=cols_time[i],
                label=f"t={t:1.2f} s",
            )
        set_axis_sci()
        plt.xlabel("rho")
        plt.ylabel("Local radiated power (W/m^3)")
        plt.legend()



import matplotlib.pylab as plt
import numpy as np
from xarray import DataArray

from indica.available_quantities import READER_QUANTITIES
from indica.converters import LineOfSightTransform
from indica.models.abstract_diagnostic import AbstractDiagnostic
from indica.numpy_typing import LabeledArray
from indica.utilities import build_dataarrays
from indica.utilities import set_axis_sci
from indica.utilities import set_plot_rcparams


class Interferometer(AbstractDiagnostic):
    """
    Object representing an interferometer diagnostics
    """

    Ne: DataArray
    los_integral_ne: DataArray

    def __init__(
        self,
        name: str,
        instrument_method="get_interferometry",
    ):
        self.transform: LineOfSightTransform
        self.name = name
        self.instrument_method = instrument_method
        self.quantities = READER_QUANTITIES[self.instrument_method]

    def _build_bckc_dictionary(self):
        bckc = {
            "t": self.t,
            "channel": np.arange(len(self.transform.x1)),
            "location": self.transform.origin,
            "direction": self.transform.direction,
            "ne": self.los_integral_ne,
        }
        self.bckc = build_dataarrays(bckc, self.quantities, transform=self.transform)

    def __call__(
        self, Ne: DataArray = None, t: LabeledArray = None, calc_rho=False, **kwargs
    ):
        """
        Calculate diagnostic measured values

        Parameters
        ----------
        Ne
            Electron density profile
        t

        Returns
        -------

        """
        # TODO: decide whether to select nearest time-points or interpolate in time!!
        if self.plasma is not None:
            if t is None:
                t = self.plasma.time_to_calculate
            Ne = self.plasma.electron_density.interp(t=t)
        else:
            if Ne is None:
                raise ValueError("Give inputs or assign plasma class!")
        self.t: DataArray = t
        self.Ne: DataArray = Ne

        los_integral_ne = self.transform.integrate_on_los(
            Ne,
            t=self.t,
            calc_rho=calc_rho,
        )
        self.los_integral_ne = los_integral_ne

        self._build_bckc_dictionary()
        return self.bckc

    def plot(self, nplot: int = 1):
        set_plot_rcparams("profiles")
        if len(self.bckc) == 0:
            print("No model results to plot")
            return

        # Line-of-sight information
        self.transform.plot()
        plt.figure()
        _value = self.bckc["ne"]
        if "beamlet" in _value.dims:
            plt.fill_between(
                _value.t,
                _value.max("beamlet"),
                _value.min("beamlet"),
                alpha=0.5,
            )
            value = _value.mean("beamlet")
        else:
            value = _value
        value.plot()
        set_axis_sci()
        plt.title(self.name.upper())
        plt.xlabel("Time (s)")
        plt.ylabel("Measured LOS-integrated density (m^-2)")
        plt.legend()



from copy import deepcopy

import matplotlib.pylab as plt
import numpy as np
from xarray import DataArray

from indica.operators.sawtooth_crash import sawtooth_crash
from indica.profilers.profiler_gauss import ProfilerGauss
from indica.utilities import set_plot_colors

CMAP, COLORS = set_plot_colors()


def density_crash(
    los_avrg=2.8e19,
    rhop=np.linspace(0, 1, 20),
    rho_inv=0.4,
    identifier="density",
):
    volume = DataArray(0.85 * rhop**3, coords=[("rhop", rhop)])

    pre = ProfilerGauss(f"electron_{identifier}", xspl=rhop)
    pre.wcenter = rho_inv / 1.5
    pre()

    plt.figure()

    scan = np.linspace(1.0, 5.0, 5)
    cols = CMAP(np.linspace(0.1, 0.75, len(scan), dtype=float))
    for i, s in enumerate(scan):
        pre.peaking = s
        pre.y0 *= los_avrg / pre.ydata.mean().values
        pre()

        post = sawtooth_crash(deepcopy(pre), rho_inv, volume)

        pre.ydata.plot(color=cols[i])
        post.ydata.plot(linestyle="dashed", color=cols[i])

    return pre, post



from abc import ABC
from abc import abstractmethod

from indica import Plasma
from indica.converters import CoordinateTransform


class AbstractDiagnostic(ABC):
    name: str = ""
    bckc: dict = {}
    plasma: Plasma
    transform: CoordinateTransform

    plasma = None

    def set_transform(self, transform: CoordinateTransform):
        """
        Line-of-sight or Transect coordinate transform
        """
        # TODO: types attribute set during initialisation!
        self.transform = transform

    def set_plasma(self, plasma: Plasma):
        """
        Assign Plasma class to use for computation of forward model
        """
        self.plasma = plasma

    def set_parameters(self, **kwargs):
        """
        Set any model kwargs
        """
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    @abstractmethod
    def _build_bckc_dictionary(self):
        """
        Calculate back-calculated expected values that the diagnostic will
        be measuring. This can be directly compared to the data read in by
        the Indica Reader methods.
        """
        raise NotImplementedError(
            "{} does not implement a "
            "'__call__' method.".format(self.__class__.__name__)
        )

    @abstractmethod
    def __call__(self, *args, **kwargs) -> dict:
        """
        Call the model and return back-calculated values
        """
        raise NotImplementedError(
            "{} does not implement a "
            "'__call__' method.".format(self.__class__.__name__)
        )



import matplotlib.pylab as plt
import numpy as np
import xarray as xr
from xarray import DataArray

from indica import Plasma
from indica.defaults.load_defaults import load_default_objects
from indica.models import BremsstrahlungDiode
from indica.models import ThomsonScattering
from indica.operators import tomo_1D
from indica.operators.tomo_1D import SXR_tomography
import indica.physics as ph
from indica.readers.modelreader import ModelReader
from indica.utilities import FIG_PATH
from indica.utilities import save_figure
from indica.utilities import set_axis_sci
from indica.utilities import set_plot_colors

CMAP, COLORS = set_plot_colors()

PLASMA = load_default_objects("st40", "plasma")
TRANSFORMS = load_default_objects("st40", "geometry")
EQUILIBRIUM = load_default_objects("st40", "equilibrium")


def example_zeff_bremstrahlung(
    plot: bool = True,
    nplot: int = 2,
    save_fig: bool = False,
):
    """
    Test workflows to calculate Zeff LOS-averaged and profile using
    Bremsstrahlung radiation measurement from passive spectroscopy.
    """

    models = {"ts": ThomsonScattering, "lines": BremsstrahlungDiode}
    TRANSFORMS["lines"] = TRANSFORMS["pi"]  # Use multi-LOS spectrometer transform
    model_reader = ModelReader(models)
    model_reader.set_plasma(PLASMA)
    model_reader.set_geometry_transforms(TRANSFORMS, EQUILIBRIUM)
    bckc = model_reader()

    filter_data = bckc["lines"]["brightness"]
    filter_data = filter_data.assign_coords(
        error=(filter_data.dims, (filter_data * 0.05).data)
    )

    print("Calculate LOS-averaged Zeff")
    zeff_los_avrg = calculate_zeff_los_averaged(
        filter_data,
        PLASMA.electron_temperature,
        PLASMA.electron_density,
        models["lines"].filter_wavelength,
    )

    print("Calculating Zeff profile from Bremss inversion (including error)")
    zeff_profile, tomo = calculate_zeff_profile(
        filter_data,
        PLASMA.electron_temperature,
        PLASMA.electron_density,
        models["lines"].filter_wavelength,
    )
    # Add phantom emissivity to tomo class for plotting purposes
    tomo.expected_emissivity = models["lines"].emissivity

    if plot:
        plot_results(
            0,
            filter_data,
            zeff_los_avrg,
            zeff_profile,
            tomo,
            nplot=nplot,
            save_fig=save_fig,
            PLASMA=PLASMA,
        )

    return zeff_los_avrg, zeff_profile, filter_data, tomo


def calculate_zeff_los_averaged(
    filter_data: DataArray,
    te_fit: DataArray,
    ne_fit: DataArray,
    filter_wavelength: float,
):
    """
    Calculate LOS-averaged Zeff any number of independent LOS

    filter_data
        Filtered diode output given the input spectra.
    te_fit
        Te profile (pon rhop)
    ne_fit
        Ne profile (pon rhop)
    filter_wavelength
        Centroid of the interference filter
    """

    los_transform = filter_data.transform
    factor_profile = ph.zeff_bremsstrahlung(
        te_fit,
        ne_fit,
        filter_wavelength,
        bremsstrahlung=xr.full_like(te_fit, 1.0),
        gaunt_approx="callahan",
    )
    factor_los_int = los_transform.integrate_on_los(factor_profile, t=filter_data.t)
    zeff_los_avrg = filter_data * factor_los_int / los_transform.los_length**2

    return zeff_los_avrg


def calculate_zeff_profile(
    filter_data: DataArray,
    te_fit: DataArray,
    ne_fit: DataArray,
    filter_wavelength: float,
):
    """
    Calculate Zeff profile from multiple LOS Bremsstrahlung brightness measurement
    using 1D inversion routine + Ne and Te profiles

    filter_data
        Filtered diode output given the input spectra.
    te_fit
        Te profile (pon rhop)
    ne_fit
        Ne profile (pon rhop)
    filter_wavelength
        Centroid of the interference filter
    """
    data = filter_data.values
    t = filter_data.t.values
    error = filter_data.error.values

    los_transform = filter_data.transform
    R = los_transform.R.mean("beamlet").values
    z = los_transform.z.mean("beamlet").values
    has_data = [True] * filter_data.shape[1]
    rho_equil = los_transform.equilibrium.rhop

    input_dict = dict(
        brightness=data,
        brightness_error=error,
        t=t,
        dl=los_transform.dl,
        R=R,
        z=z,
        rho_equil=dict(
            R=rho_equil.R.values,
            z=rho_equil.z.values,
            t=rho_equil.t.values,
            rho=rho_equil.values,
        ),
        has_data=has_data,
        debug=False,
    )
    tomo = tomo_1D.SXR_tomography(
        input_dict,
    )
    tomo_result = tomo()

    coords = [
        ("t", tomo_result["t"]),
        ("rhop", tomo_result["profile"]["rhop"][0, :]),
    ]
    emissivity = DataArray(
        tomo_result["profile"]["sym_emissivity"],
        coords=coords,
    )
    _error = DataArray(
        tomo_result["profile"]["sym_emissivity_err"],
        coords=coords,
    )
    emissivity = emissivity.assign_coords(error=(emissivity.dims, _error.data))

    wlnght = filter_wavelength
    _te = te_fit.interp(rhop=emissivity.rhop)
    _ne = ne_fit.interp(rhop=emissivity.rhop)
    zeff_profile = ph.zeff_bremsstrahlung(
        _te,
        _ne,
        wlnght,
        bremsstrahlung=emissivity,
        gaunt_approx="callahan",
    )
    zeff_lo = ph.zeff_bremsstrahlung(
        _te,
        _ne,
        wlnght,
        bremsstrahlung=emissivity - emissivity.error,
        gaunt_approx="callahan",
    )
    zeff_up = ph.zeff_bremsstrahlung(
        _te,
        _ne,
        wlnght,
        bremsstrahlung=emissivity + emissivity.error,
        gaunt_approx="callahan",
    )
    zeff_profile = zeff_profile.assign_coords(
        error=(zeff_profile.dims, np.abs(zeff_up - zeff_lo).data)
    )
    return zeff_profile, tomo


def plot_results(
    pulse: int,
    filter_data,
    zeff_los_avrg: DataArray,
    zeff_profile: DataArray,
    tomo: SXR_tomography,
    nplot: int = 2,
    save_fig: bool = False,
    fig_path: str = None,
    PLASMA: Plasma = None,
):
    los_transform = filter_data.transform

    if fig_path is None:
        fig_path = FIG_PATH

    cm, cols = set_plot_colors()

    los_transform.plot(fig_path=fig_path, fig_name=f"{pulse}_", save_fig=save_fig)

    time = zeff_profile.t
    cols = cm(np.linspace(0.1, 0.75, len(time), dtype=float))
    plt.figure()
    for i, t in enumerate(time.values):
        if i % nplot:
            continue

        zeff_profile.sel(t=t).plot(color=cols[i], label=f"t={int(t*1.e3)} ms")
        plt.fill_between(
            zeff_profile.rhop,
            (zeff_profile - zeff_profile.error).sel(t=t),
            (zeff_profile + zeff_profile.error).sel(t=t),
            color=cols[i],
            alpha=0.5,
        )
        if PLASMA is not None:
            PLASMA.zeff.sum("element").sel(t=t).plot(color=cols[i], linestyle="dashed")
    if PLASMA is not None:
        PLASMA.zeff.sum("element").sel(t=time[0]).plot(
            color=cols[0], linestyle="dashed", label="Phantom"
        )
    plt.ylabel("Zeff")
    plt.legend()
    plt.title(f"{pulse} Zeff profile")
    plt.legend()
    plt.ylim(0.5, zeff_profile.sel(rhop=slice(0, 0.5)).max())
    save_figure(fig_path, f"{pulse}_zeff_profile_PI_inversion", save_fig=save_fig)

    # Select only channels with impact parameter inside the separatrix by 1 cm
    plt.figure()
    _mean = zeff_los_avrg.mean("channel", skipna=True)
    _lo = zeff_los_avrg.min("channel", skipna=True)
    _up = zeff_los_avrg.max("channel", skipna=True)
    _mean.plot(marker="o")
    plt.fill_between(_mean.t, _lo, _up, alpha=0.5)
    plt.ylabel("Zeff")
    plt.xlabel("Time (s)")
    plt.title(f"{pulse} Zeff LOS- & channel-averaged")
    plt.ylim(
        0.5,
    )
    save_figure(fig_path, f"{pulse}_zeff_LOS_and_channel-avrg", save_fig=save_fig)

    plt.figure()
    for i, t in enumerate(time.values):
        if i % nplot:
            continue
        R = los_transform.impact_parameter.R
        to_plot = (
            filter_data.sel(t=t)
            .assign_coords(R=("channel", R.data))
            .swap_dims({"channel": "R"})
        )
        to_plot_err = filter_data.error.sel(t=t)
        to_plot.plot(color=cols[i], marker="o", label=f"t={int(t*1.e3)} ms", alpha=0.5)
        plt.errorbar(
            R,
            to_plot,
            to_plot_err,
            color=cols[i],
            alpha=0.5,
        )
        plt.scatter(
            R,
            tomo.backprojection[i, :],
            color=cols[i],
            marker="x",
            linewidths=3,
        )

    plt.scatter(
        R,
        tomo.backprojection[0, :],
        color=cols[0],
        marker="x",
        linewidths=3,
        label="Back-calculated",
    )
    plt.ylabel("Brightness (W/m$^{2}$)")
    plt.xlabel("Channel")
    plt.title("")
    plt.legend()
    set_axis_sci()
    save_figure(fig_path, f"{pulse}_bremsstrahlung_brightness", save_fig=save_fig)

    tomo.show_reconstruction()


if __name__ == "__main__":
    plt.ioff()
    example_zeff_bremstrahlung()
    plt.show()



"""Provides an abstract class defining the interface for writing out data."""

from abc import abstractmethod
from typing import Dict
from typing import Union

from xarray import DataArray
from xarray import Dataset

from indica import BaseIO
from indica import Equilibrium


class DataWriter(BaseIO):
    """An abstract class defining the interface for writing data to the
    disk or datatbases.

    """

    @abstractmethod
    def write(self, uid: str, name: str, *data: Union[Dataset, DataArray]):
        """Write data out to the desired format/database.

        The exact location will be implementation-dependent but will
        include the ``uid`` and ``name`` arguments.

        This is a wrapper function which performs tasks commons to all
        writer classes, such as converting the xarray data structures
        into a form ammenable to output. It will create a new
        :py:class:`xarray.Dataset` containing all data, with
        attributes reformated as necessary:

        - Uncertainty will be made a member of the dataset, with the name
          ``VARIABLE_uncertainty``, where ``VARIABLE`` is the name of the
          variable it is associated with.
        - Dropped data will be merged into the main data and the attribute will
          be replaced with a list of the indices of the dropped channels and
          ``dropped_dim``, the name of the dimension these indices are for.
        - The coordinate transform will be replaced with a JSON serialisation,
          from which it can be recreated. These serialisations will be stored
          in a dictionary attribute for the Dataset as a whole, with each
          DataArray holding the key for its corresponding transform.
        - Datatypes will be serialised as JSON
        - All data will have an ``equilibrium`` attribute, which provides an
          identifier for the equilibrium data (passed to the low-level writer
          in a dictionary).

        Parameters
        ----------
        uid
            User ID (i.e., user that created or wrote this data)
        name
            Name to store this data under, such as an INSTRUMENT
        data
            The data to be written out. The data will be written as though it
            had been merged into a single :py:class:`xarray.Dataset`

        """
        # TODO: Implement this

    def _write(
        self,
        uid: str,
        name: str,
        data: Dataset,
        equilibria: Dict[str, Equilibrium],
    ):
        """Perform the low-level writing of data to disk/database. It takes a
        single Dataset, with attributes reformatted as described in the
        documentation of :py:meth:`write`.

        """
        raise NotImplementedError(
            "{} does not implement a `write` method".format(self.__class__.__name__)
        )



import os
from pathlib import Path
import pickle

import corner
import flatdict
import matplotlib.pyplot as plt
import numpy as np
from skopt.plots import plot_evaluations
import xarray as xr

from indica.utilities import set_axis_sci
from indica.utilities import set_plot_rcparams


def plot_profile(
    profile: xr.DataArray,
    phantom_profile: xr.DataArray,
    label: str,
    figheader="./results/test/",
    logscale=False,
    sharefig=False,
    filename="",
    filetype=".png",
    linestyle="--",
    color="blue",
):
    set_plot_rcparams("profiles")
    plt.fill_between(
        profile.rhop,
        profile.quantile(0.16, dim="sample_idx"),
        profile.quantile(0.84, dim="sample_idx"),
        label=f"{label}, 68% Confidence",
        zorder=3,
        color=color,
        alpha=0.9,
    )
    if label != "NFAST":
        plt.fill_between(
            profile.rhop,
            profile.quantile(0.025, dim="sample_idx"),
            profile.quantile(0.975, dim="sample_idx"),
            label=f"{label}, 95% Confidence",
            zorder=2,
            color="grey",
            alpha=0.4,
        )
        plt.fill_between(
            profile.rhop,
            profile.quantile(0.005, dim="sample_idx"),
            profile.quantile(0.995, dim="sample_idx"),
            label=f"{label}, Max-Min",
            zorder=1,
            color="lightgrey",
            alpha=0.2,
        )

    if phantom_profile.any():
        phantom_profile.plot(
            label=f"{label}, phantom profile",
            linestyle=linestyle,
            color="black",
            zorder=4,
        )

    plt.legend()
    if sharefig:
        return

    set_axis_sci()
    if logscale:
        plt.yscale("log")

    plt.xlabel("Rho poloidal")
    # plt.ylabel(f"{profile.datatype[0].capitalize()} [{profile.units}]")
    if filename:
        plt.savefig(figheader + f"{filename}{filetype}")
    else:
        plt.savefig(figheader + f"{label}{filetype}")
    plt.close("all")


def _plot_1d(
    data: xr.DataArray,
    diag_data: xr.DataArray,
    filename: str,
    figheader="./results/test/",
    label="",
    ylabel="a.u.",
    xlabel="[]",
    xlim=None,
    figsize=(6.4, 4.8),
    hide_legend=False,
    capsize=3,
    markersize=4,
    elinewidth=2,
    **kwargs,
):
    set_plot_rcparams("multi")
    plt.figure(figsize=figsize)
    dims = tuple(name for name in data.dims if name != "sample_idx")
    plt.fill_between(
        data.__getattr__(dims[0]),
        data.quantile(0.16, dim="sample_idx"),
        data.quantile(0.84, dim="sample_idx"),
        label=f"{label}, 68% Confidence",
        zorder=3,
        color="blue",
    )
    plt.fill_between(
        data.__getattr__(dims[0]),
        data.quantile(0.025, dim="sample_idx"),
        data.quantile(0.975, dim="sample_idx"),
        label=f"{label}, 95% Confidence",
        zorder=2,
        color="grey",
    )
    plt.fill_between(
        data.__getattr__(dims[0]),
        data.quantile(0.005, dim="sample_idx"),
        data.quantile(0.995, dim="sample_idx"),
        label=f"{label}, Max-Min",
        zorder=1,
        color="lightgrey",
    )
    plt.errorbar(
        diag_data.__getattr__(dims[0]),
        diag_data,
        diag_data.error,
        fmt="k*",
        label=f"{label} data",
        zorder=4,
        capsize=capsize,
        markersize=markersize,
        elinewidth=elinewidth,
    )

    plt.gca().set_ylim(bottom=0)
    set_axis_sci()
    plt.ylabel(ylabel)
    plt.xlabel(xlabel)
    plt.xlim(xlim)
    if not hide_legend:
        plt.legend()
    plt.savefig(figheader + filename)
    plt.close()


def violinplot(
    data,
    diag_data,
    filename: str,
    xlabel="",
    ylabel="[a.u.]",
    figheader="./results/test/",
    **kwargs,
):
    set_plot_rcparams("multi")
    fig, axs = plt.subplots(
        1,
        1,
    )
    data = data[~np.isnan(data)]
    _data = data[((data > np.quantile(data, 0.16)) & (data < np.quantile(data, 0.84)))]

    violin = axs.violinplot(
        _data,
        showextrema=False,
        # quantiles=[0.025, 0.975, 0.16, 0.84],
        # showmedians=True,
    )
    axs.errorbar(
        1,
        y=diag_data,
        yerr=diag_data.error,
        fmt="D",
        ecolor="black",
        capsize=10,
        color="black",
    )
    violin["bodies"][0].set_edgecolor("black")
    axs.set_xlabel(xlabel)
    top = axs.get_ylim()[1]
    # bot = axs.get_ylim()[0]
    axs.set_ylim(top=top * 1.1, bottom=0)
    axs.set_ylabel(f"{ylabel}")

    set_axis_sci()
    plt.setp([axs.get_xticklabels()], visible=False)
    plt.savefig(figheader + filename)
    plt.close()


def histograms(data, diag_data, filename):
    nfig = len(data)
    fig, axs = plt.subplots(1, nfig, figsize=(16, 6))
    for idx, key in enumerate(data.keys()):
        n, bins, patches = axs[idx].hist(data[key], 50, density=True)
        q1 = (np.percentile(data[key], 16), np.percentile(data[key], 84))
        q2 = (np.percentile(data[key], 2.5), np.percentile(data[key], 97.5))
        idx_high = np.argwhere((bins > q1[0]) & (bins < q1[1])).flatten()
        idx_low = np.argwhere((bins > q2[0]) & (bins < q2[1])).flatten()
        for patch in patches:
            patch.set_facecolor("lightgrey")
        for i in idx_low:
            patches[i].set_facecolor("grey")
        for i in idx_high:
            patches[i].set_facecolor("red")

        # axs[idx].set_xlabel(f"{key} ({data[key].datatype[0]})")

        axs[idx].axvline(
            x=diag_data[key].sel(t=data[key].t).values,
            color="black",
            linestyle="-.",
            label=f"{key} data",
        )
    axs[0].set_ylabel("pdf ()")

    plt.savefig(filename)
    plt.close()


def plot_autocorr(autocorr, param_names, figheader, filetype=".png"):
    plt.figure()

    x_data = (
        np.ones(shape=(autocorr.shape))
        * np.arange(0, autocorr[:, 0].__len__())[:, None]
    )
    plt.plot(np.where(np.isfinite(autocorr), x_data, np.nan), autocorr, "x")
    plt.legend(param_names)
    plt.xlabel("iterations")
    plt.ylabel("auto-correlation time (iterations)")
    plt.savefig(figheader + "average_tau" + filetype)
    plt.close()


# flake8: noqa: C901
def plot_bda(
    results=None,
    filepath="./results/test/",
    filetype=".png",
    **kwargs,
):
    # delete all but pickle in directory and remove empty directories
    for root, dirs, files in os.walk(filepath):
        for dir in dirs:
            if not os.listdir(root + dir):
                print(f"Deleting {os.path.join(root, dir)}")
                os.rmdir(os.path.join(root, dir))
        for f in files:
            if f.endswith(".pkl"):
                continue
            else:
                print(f"Deleting {os.path.join(root, f)}")
                os.remove(os.path.join(root, f))

    if results is None:
        with open(filepath + "results.pkl", "rb") as handle:
            results = pickle.load(handle)

    # Create time directories
    time = results["TIME"]
    element = results["ELEMENT"]
    for t in time:
        Path(filepath + f"/t:{t:.2f}").mkdir(parents=True, exist_ok=True)

    diag_data = flatdict.FlatDict(results["DIAG_DATA"], ".")
    model_data = flatdict.FlatDict(results["MODEL_DATA"], ".")
    profiles = flatdict.FlatDict(results["PROFILE_STAT"]["PSI_NORM"], ".")
    post_sample = results["OPTIMISATION"]["POST_SAMPLE"]
    prior_sample = results["OPTIMISATION"]["PRIOR_SAMPLE"]
    param_names = results["OPTIMISATION"]["PARAM_NAMES"]
    phantom_profiles = flatdict.FlatDict(results["PHANTOMS"]["PSI_NORM"], ".")

    gp_regression = results["OPTIMISATION"]["CONVERGENCE"].get("gp_regression", {})
    auto_corr = results["OPTIMISATION"]["CONVERGENCE"].get("auto_corr", {})

    # select time sample_idx for plotting
    for t_idx, t in enumerate(time):
        figheader = filepath + f"t:{t:.2f}/"

        if any(gp_regression):
            plot_evaluations(gp_regression[t_idx], dimensions=param_names)
            plt.savefig(figheader + "gp_evaluations")
            plt.close()

        if any(auto_corr):
            plot_autocorr(
                auto_corr[
                    t_idx,
                ],
                param_names,
                figheader,
                filetype=filetype,
            )
        key = "EFIT.WP"
        if key in model_data.keys():
            violinplot(
                model_data[key].sel(t=t),
                diag_data[key].sel(t=t),
                f"{key.replace('.', '_')}" + filetype,
                xlabel=key,
                figheader=figheader,
                ylabel="Energy [J]",
            )

        key = "XRCS.spectra_raw"
        if key in model_data.keys():
            _plot_1d(
                model_data[key].sel(t=t),
                diag_data[key].sel(t=t),
                f"{key.replace('.', '_')}" + filetype,
                label=key,
                figheader=figheader,
                ylabel="Intensity [count/s]",
                xlabel="Wavelength [nm]",
                xlim=(0.394, 0.401),
                figsize=(15, 6),
                elinewidth=0,
                capsize=0,
            )

        cxff_quantities = [key for key in model_data.keys() if "CXFF" in key]
        for key in cxff_quantities:
            if "TI" in key:
                ylabel = "Temperature [eV]"
            elif "VTOR" in key:
                ylabel = "Toroidal Velocity [m/s]"
            else:
                ylabel = "[None]"
            _plot_1d(
                model_data[key].sel(t=t),
                diag_data[key].sel(t=t),
                f"{key.replace('.', '_')}" + filetype,
                label=key,
                figheader=figheader,
                ylabel=ylabel,
                xlabel="Channel",
                # hide_legend=True,
            )

        key = "TS.TE"
        if key in model_data.keys():
            _plot_1d(
                model_data[key].sel(t=t),
                diag_data[key].sel(t=t),
                f"{key.replace('.', '_')}" + filetype,
                label=key,
                figheader=figheader,
                ylabel="Temperature [eV]",
                xlabel="Channel",
            )
        key = "TS.NE"
        if key in model_data.keys():
            _plot_1d(
                model_data[key].sel(t=t),
                diag_data[key].sel(t=t),
                f"{key.replace('.', '_')}" + filetype,
                label=key,
                figheader=figheader,
                ylabel="Density [m^-3]",
                xlabel="Channel",
            )

        key = "TE"
        plot_profile(
            profiles[key].sel(t=t),
            phantom_profiles[key].sel(t=t),
            key,
            figheader=figheader,
            filetype=filetype,
            sharefig=True,
            color="blue",
            linestyle="dashdot",
        )
        key = "TI"
        plot_profile(
            profiles[key].sel(
                t=t,
            ),
            phantom_profiles[key].sel(
                t=t,
            ),
            key,
            figheader=figheader,
            filename="temperature",
            filetype=filetype,
            color="red",
            linestyle="dotted",
        )

        key = "VTOR"
        plot_profile(
            profiles[key].sel(
                t=t,
            ),
            phantom_profiles[key].sel(
                t=t,
            ),
            key,
            figheader=figheader,
            filename="toroidal_rotation",
            filetype=filetype,
            color="blue",
            linestyle="dotted",
        )

        key = "NE"
        plot_profile(
            profiles[key].sel(t=t),
            phantom_profiles[key].sel(t=t),
            key,
            figheader=figheader,
            filetype=filetype,
            color="blue",
            sharefig=True,
        )
        key = "NI"
        plot_profile(
            profiles[key].sel(t=t, element=element[0]),
            phantom_profiles[key].sel(t=t, element=element[0]),
            key,
            figheader=figheader,
            filetype=filetype,
            sharefig=True,
            color="red",
        )
        key = "NFAST"
        plot_profile(
            profiles[key].sel(t=t),
            phantom_profiles[key].sel(t=t),
            key,
            figheader=figheader,
            filename="densities",
            filetype=filetype,
            color="green",
        )

        key = "NI"
        for elem in element[1:]:
            plot_profile(
                profiles[key].sel(t=t, element=elem),
                phantom_profiles[key].sel(t=t, element=elem),
                key,
                figheader=figheader,
                filename=f"{elem} density",
                filetype=filetype,
                color="red",
            )

        key = "NNEUTR"
        plot_profile(
            profiles[key].sel(t=t),
            phantom_profiles[key].sel(t=t),
            key,
            filename="neutral density",
            figheader=figheader,
            filetype=filetype,
            logscale=True,
        )

        post_sample_filtered = post_sample[t_idx,][
            ~np.isnan(
                post_sample[
                    t_idx,
                ]
            ).any(axis=1)
        ]
        corner.corner(post_sample_filtered, labels=param_names)
        plt.savefig(figheader + "posterior" + filetype)

        corner.corner(
            prior_sample[
                t_idx,
            ],
            labels=param_names,
        )
        plt.savefig(figheader + "prior" + filetype)
        plt.close("all")



from matplotlib import pyplot as plt
import matplotlib.cm as cm
import numpy as np

from indica.converters import LineOfSightTransform
from indica.examples.example_plasma import example_plasma
from indica.readers.read_st40 import ReadST40


# Add Indica to python path
# path_to_indica = '/home/jonathan.wood/git_home/Indica/'
# sys.path.append(path_to_indica)

# Import Indica things


# Dummy line-of-sight
machine_dims = ((0.15, 0.85), (-0.75, 0.75))
origin_x = np.array([1.0], dtype=float)
origin_y = np.array([0.0], dtype=float)
origin_z = np.array([0.0], dtype=float)
direction_x = np.array([-1.0], dtype=float)
direction_y = np.array([0.0], dtype=float)
direction_z = np.array([0.0], dtype=float)
name = "wowzers"

origin_x = np.array([1.0], dtype=float)
origin_y = np.array([0.0], dtype=float)
origin_z = np.array([0.0], dtype=float)
direction_x = np.array([-0.8], dtype=float)
direction_y = np.array([0.4], dtype=float)
direction_z = np.array([0.0], dtype=float)
name = "wowzers"

origin_x = np.array([1.0, 1.0], dtype=float)
origin_y = np.array([0.0, 0.0], dtype=float)
origin_z = np.array([0.0, 0.0], dtype=float)
direction_x = np.array([-0.8, -0.8], dtype=float)
direction_y = np.array([0.4, 0.1], dtype=float)
direction_z = np.array([0.0, 0.0], dtype=float)
name = "wowzers"

origin_x = np.array([1.0, 1.0, 1.0], dtype=float)
origin_y = np.array([0.0, 0.0, 0.0], dtype=float)
origin_z = np.array([0.0, 0.0, 0.0], dtype=float)
direction_x = np.array([-0.8, -0.8, -0.8], dtype=float)
direction_y = np.array([0.4, 0.1, 0.0], dtype=float)
direction_z = np.array([0.0, 0.0, 0.0], dtype=float)
name = "wowzers"


beamlets = int(3 * 3)
spot_width = 0.01
# spot_height = 0.01
spot_shape = "round"
div_width = 70 * 1e-3  # radians
# div_height = 70 * 1e-3  # radians

los_transform = LineOfSightTransform(
    origin_x,
    origin_y,
    origin_z,
    direction_x,
    direction_y,
    direction_z,
    name=name,
    dl=0.01,
    spot_width=spot_width,
    # spot_height=spot_height,
    spot_shape=spot_shape,
    beamlets=beamlets,
    # div_height=div_height,
    div_width=div_width,
    machine_dimensions=machine_dims,
    passes=1,
)

# Set spot weightings
# spot_weights = SpotWeightings(
#     los_transform, "gaussian", sigma_w=0.003, sigma_v=0.003, p_w=2.0, p_v=2.0
# )
#
# los_transform.set_weightings(spot_weights.weightings)
# print(los_transform.weightings)


# Plotting...
cols = cm.gnuplot2(np.linspace(0.3, 0.75, len(los_transform.x1), dtype=float))

plt.figure()

th = np.linspace(0, 2 * np.pi, 1000)
x_ivc = machine_dims[0][1] * np.cos(th)
y_ivc = machine_dims[0][1] * np.sin(th)
x_cc = machine_dims[0][0] * np.cos(th)
y_cc = machine_dims[0][0] * np.sin(th)

plt.plot(x_cc, y_cc, c="k", lw=2.0)
plt.plot(x_ivc, y_ivc, c="k", lw=2.0)

for x1 in los_transform.x1:
    for beamlet in range(los_transform.beamlets):
        x = los_transform.x.sel(channel=x1, beamlet=beamlet)
        y = los_transform.y.sel(channel=x1, beamlet=beamlet)

        plt.plot(x, y, c=cols[x1])

plt.tight_layout()

plt.figure()

plt.plot(
    [machine_dims[0][1], machine_dims[0][1]],
    [machine_dims[1][0], machine_dims[1][1]],
    c="k",
    lw=2.0,
)

plt.plot(
    [machine_dims[0][0], machine_dims[0][0]],
    [machine_dims[1][0], machine_dims[1][1]],
    c="k",
    lw=2.0,
)

for x1 in los_transform.x1:
    for beamlet in range(los_transform.beamlets):
        R = los_transform.R.sel(channel=x1, beamlet=beamlet)
        z = los_transform.z.sel(channel=x1, beamlet=beamlet)

        plt.plot(R, z, c=cols[x1])

plt.tight_layout()
# plt.show(block=True)

"""
NEXT UP PLASMA
"""

# Inputs -- default tests equilibrium
run = "RUN01"
pulse = 10009
tstart = 0.05
tend = 0.06
dt = 0.01

# Read ST40 data
st40 = ReadST40(tstart=tstart, tend=tend, dt=dt, pulse=pulse)
st40()

# Get plasma
plasma = example_plasma(tstart=tstart, tend=tend, dt=dt, pulse=pulse)
plasma.set_equilibrium(st40.equilibrium)
los_transform.set_equilibrium(plasma.equilibrium)

#
time = plasma.t

# Along LOS
Ti = plasma.ion_temperature.sel(element="c")
print(Ti)
Ti_along_los = los_transform.map_profile_to_los(Ti, t=time)
print(Ti_along_los)

for i_time in range(len(time)):
    plt.figure()
    for i_channel in range(len(los_transform.x1)):
        for i_beamlet in range(los_transform.beamlets):
            plt.plot(
                Ti_along_los.los_position,
                Ti_along_los.sel(channel=i_channel, beamlet=i_beamlet, t=time[i_time]),
                c=cols[i_channel],
            )

    plt.ylabel("Ti (eV)")

plt.show(block=True)


# Line integration
rho = los_transform.equilibrium.rhop.interp(t=time)
R = rho.R
z = rho.z
b_tot, t = plasma.equilibrium.Btot(R, z, t=time)
b_tot_los_int = los_transform.integrate_on_los(b_tot, t=time)

b_tot_los_int_beamlets = los_transform.integrate_on_los(
    b_tot, t=time, sum_beamlets=False
)
print(b_tot_los_int)
print(b_tot_los_int_beamlets)

plt.figure()
plt.contourf(R, z, b_tot.sel(t=time[0]))
plt.colorbar()

plt.figure()
for i_beamlet in range(los_transform.beamlets):
    plt.plot(b_tot_los_int_beamlets.sel(channel=0, beamlet=i_beamlet))

plt.figure()
for i_channel in range(len(los_transform.x1)):
    plt.plot(b_tot_los_int.sel(channel=i_channel))


plt.show(block=True)



import logging
from pathlib import Path
import pickle
import pprint

import flatdict
import hydra
from omegaconf import DictConfig
from omegaconf import OmegaConf

from indica.defaults.load_defaults import load_default_objects
from indica.models import ChargeExchangeSpectrometer
from indica.models import EquilibriumReconstruction
from indica.models import HelikeSpectrometer
from indica.models import Interferometer
from indica.models import ThomsonScattering
from indica.plasma import Plasma
from indica.plasma import PlasmaProfiler
from indica.plotters.plot_bda import plot_bda
from indica.profilers.profiler_gauss import ProfilerGauss
from indica.profilers.profiler_spline import ProfilerCubicSpline
from indica.profilers.profiler_spline import ProfilerMonoSpline
from indica.readers import ReaderProcessor
from indica.readers.modelreader import ModelReader
from indica.workflows.bda.bda_driver import BDADriver
from indica.workflows.bda.optimisers import BOOptimiser
from indica.workflows.bda.optimisers import BOSettings
from indica.workflows.bda.optimisers import EmceeOptimiser
from indica.workflows.bda.optimisers import EmceeSettings
from indica.workflows.bda.priors import PriorManager

ERROR_FUNCTIONS = {
    "efit.wp": lambda x: x * 0.10,
    "xrcs.spectra_raw": lambda x: x * 0.05
    + 0.01 * x.max(dim="wavelength")
    + (
        x.where(
            (x.wavelength < 0.392) & (x.wavelength > 0.388),
        ).std("wavelength")
    ).fillna(0),
    "cxff_pi.ti": lambda x: x * 0 + 0.20 * x.max(dim="channel"),
    "cxff_tws_c.ti": lambda x: x * 0 + 0.10 * x.max(dim="channel"),
    "cxff_tws_b.ti": lambda x: x * 0 + 0.10 * x.max(dim="channel"),
}

INSTRUMENT_MAPPING: dict = {
    "xrcs": HelikeSpectrometer,
    "cxff_pi": ChargeExchangeSpectrometer,
    "cxff_tws_c": ChargeExchangeSpectrometer,
    "cxff_tws_b": ChargeExchangeSpectrometer,
    "smmh1": Interferometer,
    "efit": EquilibriumReconstruction,
    "ts": ThomsonScattering,
}


def initialise_profilers(
    x_coord, profiler_types: dict, profile_names: list, profile_params: dict = None
):
    profilers = {
        "gauss": ProfilerGauss,
        "mono_spline": ProfilerMonoSpline,
        "cubic_spline": ProfilerCubicSpline,
    }

    if profile_params is None:
        profile_params = {}
    _profilers = {
        profile_name: profilers[profiler_types[profile_name]](
            datatype=profile_name.split(":")[0],
            xspl=x_coord,
            parameters=profile_params.get(profile_name, {}),
        )
        for profile_name in profile_names
    }
    return _profilers


def save_pickle(result, filepath):
    Path(filepath).mkdir(parents=True, exist_ok=True)
    with open(filepath + "results.pkl", "wb") as handle:
        pickle.dump(result, handle)


def add_error_to_opt_data(opt_data: dict, error_functions=None, verbose=True):
    if error_functions is None:
        error_functions = ERROR_FUNCTIONS

    opt_data_with_error = {}
    for key, value in opt_data.items():
        if key not in error_functions.keys():
            opt_data_with_error[key] = value
            if verbose:
                print(f"no error function defined for {key}")
            continue
        error = error_functions[key](value)
        opt_data_with_error[key] = value.assign_coords({"error": error})
    return opt_data_with_error


@hydra.main(
    version_base=None,
    config_path="../configs/workflows/bda/",
    config_name="ion_temperature_phantom_run",
)
def bda_phantom_optimisation(  # noqa: C901
    cfg: DictConfig,
    save_results: bool = False,
):
    """
    This workflow takes all of its inputs from a config file created using hydra.
    It uses phantom data and diagnostic models to prepare an optimisation of synthetic
    data. Replace the phantom data with real experimental data to "easily" analyse a
    real scenario.

    It initialises the following objects (some of which are passed to the BDADriver):
        plasma
        profilers
        plasma_profiler
        model_coordinator
        prior_manager
        optimiser_context
    """

    log = logging.getLogger(__name__)
    config = pprint.pformat(dict(cfg))
    log.info("Beginning BDA phantom optimisation")
    log.info("Using a mock equilibrium")
    equilibrium = load_default_objects("st40", "equilibrium")

    log.info("Initialising plasma")
    plasma = Plasma(
        tstart=cfg.tstart,
        tend=cfg.tend,
        dt=cfg.dt,
        **cfg.plasma.settings,
    )
    plasma.build_atomic_data()
    plasma.set_equilibrium(equilibrium=equilibrium)

    log.info("Initialising plasma state with PlasmaProfiler")
    profilers = initialise_profilers(
        plasma.rhop,
        profiler_types=cfg.plasma.profiles.profilers,
        profile_names=cfg.plasma.profiles.params.keys(),
        profile_params=OmegaConf.to_container(cfg.plasma.profiles.params),
    )
    plasma_profiler = PlasmaProfiler(
        plasma=plasma,
        profilers=profilers,
    )
    plasma_profiler()
    log.info("Saving plasma phantom attributes")
    plasma_profiler.save_phantoms(phantom=True)

    log.info("Updating Plasma Profiler with profilers used for optimisation")
    profilers = initialise_profilers(
        plasma.rhop,
        profiler_types=cfg.plasma_profiler.profilers,
        profile_names=cfg.plasma_profiler.params.keys(),
        profile_params=OmegaConf.to_container(cfg.plasma_profiler.params),
    )
    plasma_profiler.update_profilers(profilers=profilers)

    log.info("Loading transforms from pickle")
    transforms = load_default_objects("st40", "geometry")

    log.info("Using phantom data reader")
    diagnostic_models = {diag: INSTRUMENT_MAPPING[diag] for diag in cfg.diagnostics}
    reader = ModelReader(
        models=diagnostic_models,
        model_kwargs=OmegaConf.to_container(cfg.model),
    )
    reader.set_geometry_transforms(transforms, equilibrium)
    reader.set_plasma(plasma)
    bckc = reader(
        cfg.diagnostics,
    )
    rp = ReaderProcessor(conf=cfg.reader.filters)
    processed_bckc = rp(bckc, tstart=cfg.tstart, tend=cfg.tend, dt=cfg.dt)

    flat_data = flatdict.FlatDict(processed_bckc, ".")
    log.info("Applying error to opt_data")
    opt_data = add_error_to_opt_data(flat_data, verbose=False)

    if "xrcs" in bckc.keys():
        model_call_kwargs = {
            "xrcs": {
                "norm_spectra": bckc["xrcs"]["spectra_raw"].max("wavelength"),
            }
        }
    else:
        model_call_kwargs = {}

    log.info("initialising PriorManager")
    prior_manager = PriorManager(**cfg.priors)

    if cfg.optimisation.method == "emcee":
        optimiser_settings = EmceeSettings(
            param_names=OmegaConf.to_container(cfg.param_names),
            nwalkers=cfg.optimisation.nwalkers,
            iterations=cfg.optimisation.iterations,
            sample_method=cfg.optimisation.sample_method,
            starting_samples=cfg.optimisation.starting_samples,
            burn_frac=cfg.optimisation.burn_frac,
            stopping_criteria=cfg.optimisation.stopping_criteria,
            stopping_criteria_factor=cfg.optimisation.stopping_criteria_factor,
            stopping_criteria_debug=True,
        )

        log.info("Initialising Ecmee Optimiser Context")
        optimiser_context = EmceeOptimiser(
            optimiser_settings=optimiser_settings,
            prior_manager=prior_manager,
            model_kwargs=model_call_kwargs,
        )
    elif cfg.optimisation.method == "bo":
        optimiser_settings = BOSettings(
            param_names=OmegaConf.to_container(cfg.param_names),
            n_calls=cfg.optimisation.n_calls,
            n_initial_points=cfg.optimisation.n_initial_points,
            acq_func=cfg.optimisation.acq_func,
            xi=cfg.optimisation.xi,
            noise=cfg.optimisation.noise,
            initial_point_generator=cfg.optimisation.initial_point_generator,
            use_previous_best=cfg.optimisation.use_previous_best,
            model_samples=cfg.optimisation.model_samples,
            boundary_samples=cfg.optimisation.boundary_samples,
            posterior_samples=cfg.optimisation.posterior_samples,
        )
        log.info("Initialising BO Optimiser Context")
        optimiser_context = BOOptimiser(
            optimiser_settings=optimiser_settings,
            prior_manager=prior_manager,
            model_kwargs=model_call_kwargs,
        )
    else:
        raise ValueError(
            f"cfg.optimisation.method: {cfg.optimisation.method} not implemented"
        )

    log.info("Initialising BDA Driver")
    driver = BDADriver(
        quant_to_optimise=cfg.quant_to_optimise,
        opt_data=opt_data,
        optimiser_context=optimiser_context,
        plasma_profiler=plasma_profiler,
        modelreader=reader,
        prior_manager=prior_manager,
    )

    log.info("Running BDA Driver")
    _results = driver(config=config)
    if save_results:
        filepath = "./results/phantom/"
        save_pickle(_results, filepath)
        plot_bda(_results, filepath)


if __name__ == "__main__":
    bda_phantom_optimisation()



import matplotlib.pylab as plt
import numpy as np

from indica.defaults.load_defaults import load_default_objects
from indica.models import ChargeExchangeSpectrometer
from indica.models import HelikeSpectrometer
from indica.models import PinholeCamera
from indica.models import ThomsonScattering
from indica.readers.modelreader import ModelReader
from indica.utilities import set_axis_sci
from indica.utilities import set_plot_colors

CMAP, COLORS = set_plot_colors()


def example_model_reader(plot=False):
    plasma = load_default_objects("st40", "plasma")
    transforms = load_default_objects("st40", "geometry")
    equilibrium = load_default_objects("st40", "equilibrium")
    models = {
        "cxff_pi": ChargeExchangeSpectrometer,
        "xrcs": HelikeSpectrometer,
        "ts": ThomsonScattering,
        "sxrc_xy1": PinholeCamera,
    }

    model_reader = ModelReader(models)
    model_reader.set_plasma(plasma)
    model_reader.set_geometry_transforms(transforms, equilibrium)
    bckc = model_reader()

    if plot:
        plt.ioff()
        plt.figure()
        ti = bckc["cxff_pi"]["ti"]
        cols = CMAP(np.linspace(0.75, 0.1, len(ti.t), dtype=float))
        for i, t in enumerate(ti.t):
            plasma.ion_temperature.sel(t=t, method="nearest").plot(color=cols[i])
        plasma.ion_temperature.sel(t=t, method="nearest").plot(
            color=cols[i], label="Plasma Ti"
        )
        for i, t in enumerate(ti.t):
            plt.plot(ti.transform.rhop.sel(t=t), ti.sel(t=t), "o", color=cols[i])
        plt.plot(
            ti.transform.rhop.sel(t=t),
            ti.sel(t=t),
            "o",
            color=cols[i],
            label="CXRS measurement",
        )
        plt.title("CXRS ion temperature")
        plt.legend()
        set_axis_sci()
        ti.transform.plot()
        plt.show(block=True)

        plt.figure()
        spectra = bckc["xrcs"]["spectra_raw"]
        cols = CMAP(np.linspace(0.75, 0.1, len(spectra.t), dtype=float))
        for i, t in enumerate(spectra.t):
            plt.plot(spectra.wavelength, spectra.sel(t=t), color=cols[i])
        plt.title("XRCS spectra")
        plt.legend()
        set_axis_sci()

        spectra.transform.plot()
        plt.show(block=True)

    return bckc, model_reader



from typing import Callable

import matplotlib.pylab as plt

from indica.defaults.load_defaults import load_default_objects
from indica.models import BremsstrahlungDiode
from indica.models import ChargeExchangeSpectrometer
from indica.models import EquilibriumReconstruction
from indica.models import HelikeSpectrometer
from indica.models import Interferometer
from indica.models import PinholeCamera
from indica.models import ThomsonScattering


def run_example_diagnostic_model(
    machine: str, instrument: str, model: Callable, plot: bool = False
):
    transforms = load_default_objects(machine, "geometry")
    equilibrium = load_default_objects(machine, "equilibrium")
    plasma = load_default_objects(machine, "plasma")

    plasma.set_equilibrium(equilibrium)
    transform = transforms[instrument]
    transform.set_equilibrium(equilibrium)
    model = model(instrument)
    model.set_transform(transform)
    model.set_plasma(plasma)

    bckc = model(sum_beamlets=False)

    if plot and hasattr(model, "plot"):
        plt.ioff()
        model.plot()
        plt.show()

    return plasma, model, bckc


def example_thomson_scattering(
    plot=False,
):
    machine = "st40"
    instrument = "ts"
    _model = ThomsonScattering
    return run_example_diagnostic_model(machine, instrument, _model, plot=plot)


def example_bolometer(
    plot=False,
):
    machine = "st40"
    instrument = "blom_xy1"
    _model = PinholeCamera
    return run_example_diagnostic_model(machine, instrument, _model, plot=plot)


def example_axuv_unfiltered(
    plot=False,
):
    machine = "st40"
    instrument = "sxrc_xy1"
    _model = PinholeCamera
    return run_example_diagnostic_model(machine, instrument, _model, plot=plot)


def example_charge_exchange(
    plot=False,
):
    machine = "st40"
    instrument = "cxff_pi"
    _model = ChargeExchangeSpectrometer
    return run_example_diagnostic_model(machine, instrument, _model, plot=plot)


def example_helike_spectroscopy(
    plot=False,
):
    machine = "st40"
    instrument = "xrcs"
    _model = HelikeSpectrometer
    return run_example_diagnostic_model(machine, instrument, _model, plot=plot)


def example_interferometer(
    plot=False,
):
    machine = "st40"
    instrument = "smmh"
    _model = Interferometer
    return run_example_diagnostic_model(machine, instrument, _model, plot=plot)


def example_diode_filter(
    plot=False,
):
    machine = "st40"
    instrument = "brems"
    _model = BremsstrahlungDiode
    return run_example_diagnostic_model(machine, instrument, _model, plot=plot)


def example_equilibrium(
    plot=False,
):
    machine = "st40"
    instrument = "efit"
    _model = EquilibriumReconstruction
    return run_example_diagnostic_model(machine, instrument, _model, plot=plot)


if __name__ == "__main__":
    example_helike_spectroscopy(plot=True)



from matplotlib import cm
import matplotlib.pylab as plt
import numpy as np
from xarray import DataArray

from indica.available_quantities import READER_QUANTITIES
from indica.converters import TransectCoordinates
from indica.models.abstract_diagnostic import AbstractDiagnostic
from indica.numpy_typing import LabeledArray
from indica.utilities import build_dataarrays
from indica.utilities import set_plot_rcparams


class ChargeExchangeSpectrometer(AbstractDiagnostic):
    """
    Object representing a CXRS diagnostic
    """

    def __init__(
        self,
        name: str,
        element: str = "c",
        instrument_method="get_charge_exchange",
    ):
        self.transform: TransectCoordinates
        self.name = name
        self.element = element
        self.instrument_method = instrument_method
        self.quantities = READER_QUANTITIES[self.instrument_method]

    def _build_bckc_dictionary(self):
        self.bckc = {}
        bckc = {
            "t": self.t,
            "channel": np.arange(len(self.transform.x1)),
            "x": self.transform.x,
            "y": self.transform.y,
            "z": self.transform.y,
            "R": self.transform.R,
            "vtor": self.Vtor_at_channels,
            "ti": self.Ti_at_channels,
        }
        self.bckc = build_dataarrays(bckc, self.quantities, transform=self.transform)

    def __call__(
        self,
        Ti: DataArray = None,
        Vtor: DataArray = None,
        t: LabeledArray = None,
        calc_rho: bool = False,
        **kwargs,
    ):
        """
        Calculate diagnostic measured values

        Parameters
        ----------
        Ti
            Ion temperature profile (dims = "rho", "t")
        Vtor
            Toroidal rotation profile (dims = "rho", "t")

        Returns
        -------
        Dictionary of back-calculated quantities (identical to datareader.py)

        """
        if self.plasma is not None:
            if t is None:
                t = self.plasma.time_to_calculate
            Ti = self.plasma.ion_temperature.interp(t=t)
            Vtor = self.plasma.toroidal_rotation.interp(t=t)
        else:
            if Ti is None or Vtor is None:
                raise ValueError("Give inputs or assign plasma class!")

        self.t = t
        self.Vtor = Vtor
        self.Ti = Ti

        Ti_at_channels = self.transform.map_profile_to_rho(Ti, t=t, calc_rho=calc_rho)
        Vtor_at_channels = self.transform.map_profile_to_rho(
            Vtor, t=t, calc_rho=calc_rho
        )

        self.Ti_at_channels = Ti_at_channels
        self.Vtor_at_channels = Vtor_at_channels

        self._build_bckc_dictionary()

        return self.bckc

    def plot(self, nplot: int = 1):
        set_plot_rcparams("profiles")

        cols_time = cm.gnuplot2(np.linspace(0.1, 0.75, len(self.t), dtype=float))

        self.transform.plot()

        # Plot back-calculated profiles
        plt.figure()
        for i, t in enumerate(self.t):
            if i % nplot:
                continue
            Vtor = self.bckc["vtor"].sel(t=t, method="nearest")
            rho = Vtor.transform.rhop.sel(t=t, method="nearest")
            plt.scatter(
                rho,
                Vtor,
                color=cols_time[i],
                marker="o",
                alpha=0.7,
                label=f"t={t:1.2f} s",
            )
        plt.xlabel("Channel")
        plt.ylabel("Measured toroidal rotation (rad/s)")
        plt.legend()

        plt.figure()
        for i, t in enumerate(self.t):
            if i % nplot:
                continue
            Ti = self.bckc["ti"].sel(t=t, method="nearest")
            rho = Ti.transform.rhop.sel(t=t, method="nearest")
            plt.scatter(rho, Ti, color=cols_time[i], marker="o", alpha=0.7)
        plt.xlabel("Channel")
        plt.ylabel("Measured ion temperature (eV)")
        plt.legend()
        plt.show()



from indica.models.charge_exchange_spectrometer import ChargeExchangeSpectrometer
from indica.models.diode_filters import BremsstrahlungDiode
from indica.models.equilibrium_reconstruction import EquilibriumReconstruction
from indica.models.helike_spectrometer import HelikeSpectrometer
from indica.models.interferometer import Interferometer
from indica.models.pinhole_camera import PinholeCamera
from indica.models.thomson_scattering import ThomsonScattering

MODELS_METHODS = {
    "get_interferometry": Interferometer,
    "get_charge_exchange": ChargeExchangeSpectrometer,
    # "get_spectrometer": PassiveSpectrometer,
    "get_radiation": PinholeCamera,
    "get_thomson_scattering": ThomsonScattering,
    "get_helike_spectroscopy": HelikeSpectrometer,
    "get_diode_filters": BremsstrahlungDiode,
    # "get_spectrometer": PassiveSpectrometer,
}

__all__ = [
    "PinholeCamera",
    "ChargeExchangeSpectrometer",
    "BremsstrahlungDiode",
    "EquilibriumReconstruction",
    "HelikeSpectrometer",
    "Interferometer",
    "ThomsonScattering",
]



"""Module to provide objects for reading fusion data from the
filesystem or databases. These objects are all subclasses of
:py:class:`readers.DataReader`, with each subclass providing
functionality for a different format of data.

"""
from .adas import ADASReader
from .datareader import DataReader
from .readerprocessor import ReaderProcessor

__all__ = [
    "ADASReader",
    "DataReader",
    "ReaderProcessor",
]

try:
    from .salutils import SALUtils

    __all__ += ["SALUtils"]
except ImportError:
    pass

try:
    from .jetreader import JETReader

    __all__ += ["JETReader"]
except ImportError:
    pass


try:
    from .mdsutils import MDSUtils

    __all__ += ["MDSUtils"]
except ImportError as e:
    print(e)
    pass


try:
    from .st40reader import ST40Reader

    __all__ += ["ST40Reader"]
except ImportError as e:
    print(e)
    pass



from pathlib import Path

from indica.configs.readers.machineconf import MachineConf


class JETConf(MachineConf):
    def __init__(self):
        self.MACHINE_DIMS = ((1.83, 3.9), (-1.75, 2.0))
        self.INSTRUMENT_METHODS = {
            "efit": "get_equilibrium",
            "eftp": "get_equilibrium",
            "hrts": "get_thomson_scattering",
            "lidr": "get_thomson_scattering",
            "kk3": "get_cyclotron_emissions",
            "kg10": "get_density_reflectometer",
            "sxrh": "get_radiation",
            "sxrv": "get_radiation",
            "sxrt": "get_radiation",
            "kb5h": "get_radiation",
            "kb5v": "get_radiation",
            "ks3h": "get_zeff",
            "ks3v": "get_zeff",
            **{
                "cx{}{}".format(val1, val2): "get_charge_exchange"
                for val1 in ("s", "d", "f", "g", "h")
                for val2 in ("m", "w", "x", "4", "6", "8")
            },
        }
        self.QUANTITIES_PATH = {
            "get_equilibrium": {
                "R": "psir",
                "z": "psiz",
                "rgeo": "rgeo",
                "rmag": "rmag",
                "zmag": "zmag",
                "psi_axis": "faxs",
                "psi_boundary": "fbnd",
                "wp": "wp",
                "rbnd": "rbnd",
                "zbnd": "zbnd",
                "f": "f",
                "ftor": "ftor",
                "rmji": "rmji",
                "rmjo": "rmjo",
                "vjac": "vjac",
                "ajac": "ajac",
                "psi": "psi",
            },
            "get_thomson_scattering": {
                "z": "z",
                "ne": "ne",
                "te": "te",
            },
            "get_cyclotron_emissions": {"te": "te"},
            "get_reflectometer": {"ne": "ne"},
            "get_sxr_radiation": {},
            "get_radiation": {},
            "get_zeff": {},
            "get_charge_exchange": {
                "R": "rpos",
                "z": "pos",
                "ti": "ti",
                "vtor": "angf",
            },
        }
        self._BREMSSTRAHLUNG_LOS = {
            "ks3": "edg7",
        }
        self._RADIATION_RANGES = {
            "sxr/h": 17,
            "sxr/t": 35,
            "sxr/v": 35,
            "bolo/kb5h": 24,
            "bolo/kb5v": 32,
        }
        self._KK3_RANGE = (1, 96)
        self.SURF_PATH = Path(__file__).parent.parent.parent / "data/surf_los.dat"



"""Experimental design for performing mathematical operations on data.
"""

from abc import ABC
from abc import abstractmethod
import datetime
from typing import Any
from typing import Union

from xarray import DataArray
from xarray import Dataset

Data = Union[DataArray, Dataset]


class OperatorError(Exception):
    """An Exception class raised by :py:class:`operator.Operator` when
    receiving erroneous arguments.
    """


class Operator(ABC):
    """Abstract base class for performing calculations with data."""

    def __init__(self, **kwargs: Any):
        """Creates a provenance - currently not very elaborate."""
        self._init_time: datetime.datetime

    @abstractmethod
    def __call__(self, *args: DataArray) -> Union[DataArray, Dataset]:
        """The invocation of the operator.

        The exact number of arguments should be determined by the
        subclass. However, it is anticipated that these would all be
        :py:class:`xarray.DataArray` objects.

        Unfortunately, we can not use Mypy static type-checking for
        this routine or its overriding implementations, as the number
        of arguments will vary.

        """
        raise NotImplementedError(
            "{} does not implement a "
            "'__call__' method.".format(self.__class__.__name__)
        )



from copy import deepcopy

import numpy as np
from scipy.interpolate import CubicSpline
from scipy.optimize import least_squares
import xarray as xr
from xarray import DataArray

from indica import Equilibrium
from indica.numpy_typing import ArrayLike


def fit_profile_and_R_shift(
    Rdata: DataArray,
    zdata: DataArray,
    ydata: DataArray,
    yerr: DataArray,
    equilibrium: Equilibrium,
    xknots: ArrayLike = None,
    R_shift: DataArray = None,
    xspl: ArrayLike = np.linspace(0, 1.05, 51),
    bc_type: str = "clamped",
    bounds_R_shift: tuple = (-0.02, 0.02),
    verbose: bool = False,
    default_perc_err: float = 0.05,
):
    """Fit a profile and the R_shift of the equilibrium"""

    def residuals(all_knots):
        if R_shift is None:
            _R_shift = all_knots[-1]
            yknots = all_knots[:-1]
        else:
            _R_shift = R_shift.sel(t=t).values
            yknots = all_knots

        _xknots, _, _ = equilibrium.flux_coords(R + _R_shift, z, t=t)

        spline = CubicSpline(
            xknots,
            yknots,
            axis=0,
            bc_type=bc_type,
        )
        bckc = np.interp(_xknots, xspl, spline(xspl))

        _residuals = (y - bckc) / err

        return _residuals

    # Boundary conditions
    # values go to --> 0 outside separatrix (index = -2)
    nknots = np.size(xknots)
    lower_bound = np.full(nknots, -np.inf)
    upper_bound = np.full(nknots, np.inf)
    lower_bound[-1] = 0.0
    upper_bound[-1] = 0.01
    if R_shift is None:
        # If R_shift to be fitted, add additional knot and bounds
        nknots += 1
        lower_bound = np.append(lower_bound, bounds_R_shift[0])
        upper_bound = np.append(upper_bound, bounds_R_shift[1])

    # Initialize DataArray that will contain the final fit result
    yspl = xr.DataArray(
        np.empty((len(xspl), len(ydata.t))),
        coords=[("rhop", xspl), ("t", ydata.t.values)],
    )
    R_shift_fit = xr.DataArray(
        np.empty(len(ydata.t)),
        coords=[("t", ydata.t.values)],
    )
    rho_data = xr.full_like(ydata, np.nan)

    all_knots = None
    for t in ydata.t.values:
        # Normalize data so range of parameters to scan is all similar
        norm_factor = np.nanmax(ydata.sel(t=t).values)
        _y = ydata.sel(t=t).values / norm_factor
        _yerr = yerr.sel(t=t).values / norm_factor
        if np.all(_yerr == 0):
            _yerr = _y * default_perc_err

        _R_shift = 0.0
        _yspl = np.full_like(xspl, 0.0)
        ind = np.where(np.isfinite(_y) * np.isfinite(_yerr))[0]
        if len(ind) > 2:
            R = Rdata[ind]
            z = zdata[ind]
            y = _y[ind]
            err = _yerr[ind]

            # Initial guess: profile linearly increasing edge>core & R_shift = 0.
            if all_knots is None:
                all_knots = np.linspace(np.max(y), 0, np.size(xknots))
                if R_shift is None:
                    all_knots = np.append(all_knots, 0.0)

            try:
                fit = least_squares(
                    residuals,
                    all_knots,
                    bounds=(lower_bound, upper_bound),
                    verbose=verbose,
                )

                if R_shift is None:
                    yknots = fit.x[:-1]
                    _R_shift = fit.x[-1]
                else:
                    yknots = fit.x
                    _R_shift = R_shift.sel(t=t)

                all_knots = deepcopy(fit.x)
                spline = CubicSpline(
                    xknots,
                    yknots,
                    axis=0,
                    bc_type=bc_type,
                )

                _yspl = spline(xspl) * norm_factor
            except ValueError:
                all_knots = None
                _R_shift = 0.0
                _yspl = np.full_like(xspl, 0.0)

        yspl.loc[dict(t=t)] = _yspl
        if R_shift is None:
            R_shift_fit.loc[dict(t=t)] = _R_shift

        _rho_data, _, _ = equilibrium.flux_coords(Rdata + _R_shift, zdata, t=t)
        rho_data.loc[dict(t=t)] = _rho_data

    if R_shift is not None:
        R_shift_fit = R_shift

    return yspl, R_shift_fit, rho_data



import numpy as np
from scipy.interpolate import CubicSpline
from scipy.optimize import least_squares
import xarray as xr
from xarray import DataArray

from indica.numpy_typing import ArrayLike


def fit_profile(
    xdata: DataArray,
    ydata: DataArray,
    yerr: DataArray,
    xspl: ArrayLike = np.linspace(0, 1.05, 51),
    knots: ArrayLike = None,
    virtual_knots: bool = False,
    bc_type: str = "clamped",
    verbose=0,
):
    """Fit a profile"""

    def residuals(yknots):
        spline = CubicSpline(
            xknots,
            yknots,
            axis=0,
            bc_type=bc_type,
        )
        bckc = np.interp(x, xspl, spline(xspl))
        return (y - bckc) / err

    yspl = xr.DataArray(
        np.empty((len(xspl), len(ydata.t))),
        coords=[("rhop", xspl), ("t", ydata.t.values)],
    )
    for t in ydata.t.values:
        _x = xdata.sel(t=t).values
        _y = ydata.sel(t=t).values
        _yerr = yerr.sel(t=t).values

        ind = np.where(np.isfinite(_x) * np.isfinite(_y) * np.isfinite(_yerr))[0]
        if len(ind) > 2:
            x_to_sort = _x[ind]
            y_to_sort = _y[ind]
            err_to_sort = _yerr[ind]

            ind_sort = np.argsort(x_to_sort)
            x = x_to_sort[ind_sort]
            y = y_to_sort[ind_sort]
            err = err_to_sort[ind_sort]

            if knots is None:
                xknots = np.append(np.append(0.0, x), 1.05)
            else:
                xknots = knots
            yknots = np.interp(xknots, x, y)
            yknots[-1] = 0.0

            if virtual_knots:
                ind = range(len(x) - 1)
                dx = [x[i + 1] - x[i] for i in ind]
                xknots_virt = [
                    x[i + 1] - dx[i] / 2 for i in ind if (i % 2 == 0 and x[i + 1] > 0.5)
                ]
                _xknots = np.unique(np.sort(np.append(xknots, xknots_virt)))
                yknots = np.interp(_xknots, xknots, yknots)
                xknots = _xknots

            lower_bound = np.full_like(xknots, -np.inf)
            upper_bound = np.full_like(xknots, np.inf)
            lower_bound[-1] = 0.0
            upper_bound[-1] = 0.01

            try:
                fit = least_squares(
                    residuals,
                    yknots,
                    bounds=(lower_bound, upper_bound),
                    verbose=verbose,
                )
                spline = CubicSpline(
                    xknots,
                    fit.x,
                    axis=0,
                    bc_type=bc_type,
                )
                _yspl = spline(xspl)
            except ValueError:
                _yspl = np.full_like(xspl, 0.0)
        else:
            _yspl = np.full_like(xspl, 0.0)

        yspl.loc[dict(t=t)] = _yspl

    return yspl



from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from indica.readers.read_st40 import ReadST40
from indica.utilities import FIG_PATH
from indica.utilities import save_figure
from indica.utilities import set_axis_sci
from indica.utilities import set_plot_colors
from indica.utilities import set_plot_rcparams

CMAP, COLORS = set_plot_colors()
LINESTYLES = ["solid", "dashed", "dotted"]
set_plot_rcparams("profiles")
QUANTITIES: list = [
    "smmh:ne",
    "xrcs:ti_w",
    "xrcs:te_n3w",
    "xrcs:te_kw",
    "xrcs:spectra",
    "sxrc_xy2:brightness",
    "efit:ipla",
    "efit:rmag",
    "efit:zmag",
    "sxr_spd:brightness",
    # "cxff_pi:ti",
    # "cxff_pi:vtor",
    # "cxff_tws_c:ti",
    # "cxff_tws_c:vtor",
    "ts:te",
    "ts:ne",
    # "lines",
]

Y0 = {}
Y0["nirh1"] = True
Y0["smmh1"] = True
Y0["xrcs"] = True
Y0["sxr_diode_1"] = True
Y0["brems"] = True
Y0["efit"] = True
Y0["cxff_tws_c"] = True
Y0["cxff_pi"] = True


def plot_st40_data(
    pulses: list,
    tstart: float = 0.02,
    tend: float = 0.1,
    dt: float = 0.005,
    quantities: list = [],
    tplot: float = None,
    save_fig: bool = False,
    fig_path: str = None,
    fig_style: str = "profiles",
    plot_binned: bool = True,
):

    if tplot is None:
        tplot = np.mean([tstart, tend])
    if len(quantities) == 0:
        quantities = QUANTITIES
    instruments = list(np.unique([quant.split(":")[0] for quant in quantities]))
    if fig_path is None:
        fig_path = f"{FIG_PATH}"

    set_plot_rcparams(fig_style)
    xr.set_options(keep_attrs=True)
    if len(pulses) > 1:
        colors = CMAP(np.linspace(0.75, 0.1, len(pulses), dtype=float))
    else:
        colors = ["blue"]

    raw: dict = {quant: {} for quant in quantities}
    binned = deepcopy(raw)
    data = {"raw": raw, "binned": binned}
    for i, pulse in enumerate(pulses):
        fig_path += f"_{pulse}"
        st40 = ReadST40(pulse, tstart=tstart, tend=tend, dt=dt)
        st40(instruments)
        for quantity in quantities:
            instr, quant = quantity.split(":")
            data["binned"][quantity][pulse] = st40.binned_data[instr][quant]
            data["raw"][quantity][pulse] = st40.raw_data[instr][quant]

    for quantity in quantities:
        print(quantity)
        instr, _ = quantity.split(":")
        plt.figure()
        for i, pulse in enumerate(pulses):
            color = colors[i]
            if len(data["raw"][quantity].keys()) == 0:
                continue
            plot_data(data, quantity, pulse, tplot, key="raw", color=color)
            if plot_binned:
                plot_data(data, quantity, pulse, tplot, key="binned", color=color)

            set_axis_sci()
            if instr in Y0.keys():
                plt.ylim(
                    0,
                )
            plt.legend()
            plt.autoscale()
            save_figure(fig_path, f"{quantity}", save_fig=save_fig)

    return data, st40


def plot_data(data, quantity: str, pulse: int, tplot: float, key="raw", color=None):
    str_to_add = ""
    instr, quant = quantity.split(":")
    if key == "raw":
        marker = None
    else:
        marker = "o"

    _data = data[key][quantity][pulse]
    tslice = slice(_data.t.min().values, _data.t.max().values)
    if "error" not in _data.attrs:
        _data = _data.assign_coords(error=(_data.dims, xr.full_like(_data, 0.0).data))
    if "stdev" not in _data.attrs:
        _data = _data.assign_coords(stdev=(_data.dims, xr.full_like(_data, 0.0).data))
    _err = np.sqrt(_data.error**2 + _data.stdev**2)
    _err = xr.where(_err / _data.values < 1.0, _err, 0.0)
    if len(_data.dims) > 1:
        str_to_add = f" @ {tplot:.3f} s"
        tslice = _data.t.sel(t=tplot, method="nearest")

    _data = _data.sel(t=tslice)
    _err = _err.sel(t=tslice)
    if instr in "xrcs" and quant == "spectra":
        bgnd = _data.sel(wavelength=slice(0.393, 0.388)).mean("wavelength")
        _data -= bgnd
    label = None
    if key == "raw":
        label = str(pulse)
        alpha = 0.5
        plt.fill_between(
            _data.coords[_data.dims[0]].values,
            _data.values - _err.values,
            _data.values + _err.values,
            color=color,
            alpha=alpha,
        )
    if key == "binned":
        alpha = 0.8
        plt.errorbar(
            _data.coords[_data.dims[0]].values,
            _data.values,
            _err.values,
            color=color,
            alpha=alpha,
        )
    _data.plot(label=label, color=color, alpha=alpha, marker=marker)
    plt.title(f"{instr.upper()} {quant}" + str_to_add)


if __name__ == "__main__":
    plt.ioff()
    plot_st40_data([11226, 11225])
    plt.show()



from indica.configs.readers.st40readerprocessorconf import ST40ReaderProcessorConf
from .jetconf import JETConf
from .st40conf import ST40Conf

__all__ = [
    "ST40Conf",
    "JETConf",
    "ST40ReaderProcessorConf",
]



"""
Quantities read by each reader method of the datareader class.

Dictionary structure:
    {"Name of the reader method":
        "Quantity name" :{
            ("Datatype identified",
                ["Coord1 quantity name", "Coord2  quantity name", ...])
        }

For the Datatype identifier string see indica/datatypes.py.

NB Dimensions (having Coordinates == []) must have key == datatype identifier

"""

from typing import Dict
from typing import Tuple

READER_QUANTITIES: Dict[str, Dict[str, Tuple[str, list]]] = {
    "get_thomson_scattering": {
        "t": ("t", ["t"]),
        "channel": ("channel", []),
        "x": ("x", []),
        "y": ("y", []),
        "z": ("z", []),
        "R": ("R", []),
        "ne": ("electron_density", ["t", "channel"]),
        "te": ("electron_temperature", ["t", "channel"]),
        "chi2": ("chi_squared", ["t", "channel"]),
    },
    "get_spectrometer": {
        "t": ("t", ["t"]),
        "channel": ("channel", []),
        "location": ("location", []),
        "direction": ("direction", []),
        "wavelength": ("wavelength", []),
        "spectra_raw": ("spectra_raw", ["t", "wavelength"]),
        "spectra": ("spectra", ["t", "channel", "wavelength"]),
    },
    "get_charge_exchange": {
        "t": ("t", ["t"]),
        "channel": ("channel", []),
        "wavelength": ("wavelength", []),
        "x": ("x", []),
        "y": ("y", []),
        "z": ("z", []),
        "R": ("R", []),
        "location": ("location", []),
        "direction": ("direction", []),
        "vtor": ("toroidal_rotation", ["t", "channel"]),
        "ti": ("ion_temperature", ["t", "channel"]),
        "conc": ("concentration", ["t", "channel"]),
        "spectra": ("spectra", ["t", "channel", "wavelength"]),
        "fit": ("spectra_fit", ["t", "channel", "wavelength"]),
    },
    "get_helike_spectroscopy": {
        "t": ("t", ["t"]),
        "channel": ("channel", []),
        "wavelength": ("wavelength", []),
        "location": ("location", []),
        "direction": ("direction", []),
        "ti_w": ("ion_temperature", ["t"]),
        "ti_z": ("ion_temperature", ["t"]),
        "te_n3w": ("electron_temperature", ["t"]),
        "te_kw": ("electron_temperature", ["t"]),
        "int_w": ("line_intensity", ["t"]),
        "int_k": ("line_intensity", ["t"]),
        "int_tot": ("line_intensity", ["t"]),
        "int_n3": ("line_intensity", ["t"]),
        "spectra_raw": ("spectra_raw", ["t", "wavelength"]),
        "spectra": ("spectra", ["t", "wavelength"]),
        "background": ("intensity", ["t"]),
    },
    "get_profile_fits": {
        "t": ("t", ["t"]),
        "channel": ("channel", []),
        "rhop": ("rhop", []),
        "R": ("R_fit", []),
        "z": ("z", []),
        "rhop_data": ("rhop", []),
        "R_data": ("R", []),
        "ne_rhop": ("electron_density", ["t", "rhop"]),
        "te_rhop": ("electron_temperature", ["t", "rhop"]),
        "pe_rhop": ("electron_pressure", ["t", "rhop"]),
        "ne_R": ("electron_density", ["t", "R"]),
        "te_R": ("electron_temperature", ["t", "R"]),
        "pe_R": ("electron_pressure", ["t", "R"]),
        "ne_data": ("electron_density", ["t", "channel"]),
        "te_data": ("electron_temperature", ["t", "channel"]),
        "pe_data": ("electron_pressure", ["t", "channel"]),
        "R_shift": ("R_shift", ["t"]),
    },
    "get_diode_filters": {
        "t": ("t", ["t"]),
        "channel": ("channel", []),
        "label": ("label", []),
        "location": ("location", []),
        "direction": ("direction", []),
        "brightness": ("brightness", ["t", "channel"]),
    },
    "get_interferometry": {
        "t": ("t", ["t"]),
        "channel": ("channel", []),
        "location": ("location", []),
        "direction": ("direction", []),
        "ne": ("electron_density_integrated", ["t"]),
    },
    "get_equilibrium": {
        "t": ("t", ["t"]),
        "psin": ("psin", []),
        "index": ("index", []),
        "R": ("R", []),
        "z": ("z", []),
        "rgeo": ("R_geo", ["t"]),
        "rmag": ("R_mag", ["t"]),
        "zmag": ("z_mag", ["t"]),
        "psi_axis": ("poloidal_flux_axis", ["t"]),
        "psi_boundary": ("poloidal_flux_boundary", ["t"]),
        "ipla": ("plasma_current", ["t"]),
        "wp": ("equilibrium_stored_energy", ["t"]),
        "rbnd": ("R_boundary", ["t", "index"]),
        "zbnd": ("z_boundary", ["t", "index"]),
        "f": ("equilibrium_f", ["t", "psin"]),
        "ftor": ("toroidal_flux", ["t", "psin"]),
        "rmji": ("R_hfs", ["t", "psin"]),
        "rmjo": ("R_lfs", ["t", "psin"]),
        "vjac": ("volume_jacobian", ["t", "psin"]),
        "ajac": ("area_jacobian", ["t", "psin"]),
        "psi": ("poloidal_flux", ["t", "z", "R"]),
    },
    "get_radiation": {
        "t": ("t", ["t"]),
        "channel": ("channel", []),
        "label": ("label", []),
        "location": ("location", []),
        "direction": ("direction", []),
        "brightness": ("brightness", ["t", "channel"]),
    },
    "get_zeff": {
        "t": ("t", ["t"]),
        "rhop": ("rhop", []),
        "zeff_avrg": ("effective_charge", ["t"]),
        "zeff_hi": ("effective_charge", ["t"]),
        "zeff_low": ("effective_charge", ["t"]),
        "zeff": ("effective_charge", ["t", "rhop"]),
        "R_shift": ("R_shift", ["t"]),
    },
    # "get_astra": {  # TODO: implement
    # },
}



import getpass


user = getpass.getuser()


def astra():
    psu_dict = {
        "I": ("SIGNAL", "Current, MA"),
        "V": ("SIGNAL", "Voltage, V"),
    }
    constraints_dict = {
        "INDEX": ("NUMERIC", "x vector(i) = i"),
        "CVALUE": ("SIGNAL", "simulated"),
        "MVALUE": ("SIGNAL", "experimental"),
        "WEIGHT": ("SIGNAL", ""),
    }
    nodes = {
        "CODE_VERSION": {
            "USER": ("TEXT", "User responsible of run"),
        },
        "": {
            "TIME": ("NUMERIC", "time vector, s"),
        },  # (type, description)
        ".GLOBAL": {
            "IPL": ("SIGNAL", "Plasma current, MA"),
            "BTVAC": ("SIGNAL", "BT_vacuum at R=0.5m, T"),
            "DF": ("SIGNAL", "simulated diamagnetic flux, Wb"),
            "CR0": ("SIGNAL", "MINOR RAD=(Rmax-Rmin)/2, m"),
            "RGEO": ("SIGNAL", "MAJOR R = (Rmax+Rmin)/2, m"),
            "ZGEO": ("SIGNAL", "Geom vertical position, m "),
            "RC": ("SIGNAL", "Current density center R_IP, m "),
            "ZC": ("SIGNAL", "Current density center Z_IP, m "),
            "ELON": ("SIGNAL", "ELONGATION BOUNDARY"),
            "TRIL": ("SIGNAL", "LOWER TRIANGULARITY"),
            "TRIU": ("SIGNAL", "UPPER TRIANGULARITY"),
            "QWL": ("SIGNAL", "Q(PSI) AT the LCFS"),
            "Q95": ("SIGNAL", "Q(PSI) AT 95% of full poloidal flux inside LCFS"),
            "TE0": ("SIGNAL", "Central electron temp, keV"),
            "TI0": ("SIGNAL", "Central ion temp, keV"),
            "NE0": ("SIGNAL", "Central electron density, 10E19m^-3 "),
            "NEL": ("SIGNAL", "Line aver electron density 10E19m^-3 "),
            "NEV": ("SIGNAL", "Volume aver electron density 10E19m^-3 "),
            "TEV": ("SIGNAL", "Volume aver electron temp, keV"),
            "TIV": ("SIGNAL", "Volume aver ion temp, keV"),
            "TAUE": ("SIGNAL", "Energy confinement time, s "),
            "P_OH": ("SIGNAL", "Total Ohmic power, MW"),
            "IEXC": ("SIGNAL", "Ion-electron exchange power, MW"),
            "UPL": ("SIGNAL", "Loop Voltage,V"),
            "WTH": ("SIGNAL", "Thermal energy, MJ"),
            "Li3": ("SIGNAL", "Internal inductance"),
            "BetP": ("SIGNAL", "Poloidal beta"),
            "BetT": ("SIGNAL", "Toroidal beta"),
            "BetN": ("SIGNAL", "Beta normalized  "),
            "Hoh": ("SIGNAL", "Neo-alcator H-factor"),
            "H98": ("SIGNAL", "ITER IPB(y,2) H-factor"),
            "HNSTX": ("SIGNAL", "NSTX scaling H-factor"),
            "HPB": ("SIGNAL", "Peter Buxton H-factor"),
            "ZEFF": ("SIGNAL", "Z effective at the plasma center"),
            "Res": ("SIGNAL", "Total plasma resistance Qj/Ipl^2, Ohm"),
            "Rmag": ("SIGNAL", "Magnetic axis hor position, m"),
            "Zmag": ("SIGNAL", "Magnetic axis vert position, m"),
            "Vol": ("SIGNAL", "Plasma volume, m^3"),
            "ROC": ("SIGNAL", "Effective plasma radius, m"),
            "P_NBI_E": ("SIGNAL", "Total power from NBI to electrons, MW"),
            "P_NBI_I": ("SIGNAL", "Total power from NBI to ions, MW"),
            "P_RF": ("SIGNAL", "Total RF power to electrons,MW"),
            "I_BS": ("SIGNAL", "Total bootstrap current,MA"),
            "F_BS": ("SIGNAL", "Bootstrap current fraction"),
            "I_NBI": ("SIGNAL", "Total NB driven current,MA"),
            "NBI_NAMES": ("SIGNAL", "NBI names, RFX,HNBI1=00,01,10,11"),
            "I_RF": ("SIGNAL", "Total RF driven current,MA"),
            "I_OH": ("SIGNAL", "Total Ohmic current,MA"),
            "F_NI": ("SIGNAL", "Non-inductive current fraction"),
            "P_FUS_THERM": ("SIGNAL", "Thermal fusion power,MW"),
            "P_FUS_TOT": ("SIGNAL", "Total fusion power: thermal+NBI,MW"),
            "P_AUX": ("SIGNAL", "Total external heating power,MW"),
            "Q_FUS": ("SIGNAL", "Fusion energy gain"),
            "P_TOT_E": ("SIGNAL", "Total alpha power to electrons,MW"),
            "P_TOT_I": ("SIGNAL", "Total alpha power to ions,MW"),
            "psi_boundary": ("SIGNAL", "boundary poloidal flux,Wb"),
            "FAXS": ("SIGNAL", "axis poloidal flux,Wb"),
            "QE": ("SIGNAL", "electron power flux through LCFS, MW"),
            "QI": ("SIGNAL", "ion power flux through LCFS, MW"),
            "QN": ("SIGNAL", "electron flux through LCFS, 10^19/s"),
            "STOT": ("SIGNAL", "Total electron source, 10^19/s"),
            "SPEL": ("SIGNAL", "Pellet electron source, 10^19/s"),
            "SWALL": ("SIGNAL", "Boundary electron source, 10^19/s"),
            "SBM": ("SIGNAL", "Neutral beam electron source, 10^19/s"),
            "NNCL": ("SIGNAL", "Wall cold neutral density, 10^19/m^3"),
            "TAUP": ("SIGNAL", "Particle confinement time ,s"),
            "GAMMA": ("SIGNAL", "KINX growth rate, 1/s"),
            "VTOR0": ("SIGNAL", "Central toroidal velocity, m/s"),
            "TORQ": ("SIGNAL", "Total torque from NB, N*m"),
            "TORQ_BE": ("SIGNAL", "Collisional to electron torque from NB, N*m"),
            "TORQ_BI": ("SIGNAL", "Collisional to ions torque from NB, N*m"),
            "TORQ_BTH": ("SIGNAL", "Beam thermalisation torque from NB, N*m"),
            "TORQ_JXB": ("SIGNAL", "JXB torque from NB, N*m"),
            "TORQ_BCX": ("SIGNAL", "CX losses torque from NB, N*m"),
            "TAU_PHI": ("SIGNAL", "Momentum confinement time, s"),
        },
        ".PSI2D:": {
            "RGRID": ("NUMERIC", "Major radius coordinate m"),
            "ZGRID": ("NUMERIC", "Vertical coordinate m"),
            "PSI": ("SIGNAL", "Poloidal flux W"),
        },
        ".PSU.CS": psu_dict,
        ".PSU.MC": psu_dict,
        ".PSU.DIV": psu_dict,
        ".PSU.PSH": psu_dict,
        ".PSU.BVU": psu_dict,
        ".PSU.BVUT": psu_dict,
        ".PSU.BVUB": psu_dict,
        ".PSU.BVL": psu_dict,
        ".PSU.PSH": psu_dict,
        ".CONSTRAINTS.BP": constraints_dict,
        ".CONSTRAINTS.DF": constraints_dict,
        ".CONSTRAINTS.FLUX": constraints_dict,
        ".CONSTRAINTS.IP": constraints_dict,
        ".CONSTRAINTS.PFC_DOF": constraints_dict,
        ".CONSTRAINTS.PRESSURE": constraints_dict,
        ".CONSTRAINTS.ROGC": constraints_dict,
        ".CONSTRAINTS.ULOOP": constraints_dict,
        ".PROFILES.PSI_NORM": {
            "XPSN": ("NUMERIC", "x vector -sqrt(fi_normalized)"),
            "Q": ("SIGNAL", "Q_PROFILE(PSI_NORM)"),
            "P": ("SIGNAL", "PRESSURE(PSI_NORM)"),
            "PSI": ("SIGNAL", "PSI"),
            "PPRIME": ("SIGNAL", "PPRIME"),
            "FFPRIME": ("SIGNAL", "FFPRIME"),
            "FTOR": ("SIGNAL", "Toroidal flux, Wb"),
            "SIGMAPAR": ("SIGNAL", "Parallel conductivity,1/(Ohm*m)"),
            "AREAT": ("SIGNAL", "Toroidal cross section,m2"),
            "VOLUME": ("SIGNAL", "Volume inside magnetic surface,m3"),
        },
        ".PROFILES.ASTRA": {
            "RHO": ("SIGNAL", "rho - toroidal flux coordinate, m"),
            "TE": ("SIGNAL", "Electron temperature, keV"),
            "NE": ("SIGNAL", "Electron density, 10^19 m^-3"),
            "NI": ("SIGNAL", "Main ion density, 10^19 m^-3"),
            "TI": ("SIGNAL", "Ion temperature, keV"),
            "ZEFF": ("SIGNAL", "Effective ion charge"),
            "QNBE": ("SIGNAL", "Beam power density to electrons, MW/m3"),
            "QNBI": ("SIGNAL", "Beam power density to ions, MW/m3"),
            "Q_OH": ("SIGNAL", "Ohmic heating power profile, MW/m3"),
            "STOT": ("SIGNAL", "Total electron source,10^19/s/m3"),
            "CHI_E": ("SIGNAL", "Total electron heat conductivity, m^2/s"),
            "CHI_I": ("SIGNAL", "Total ion heat conductivity, m^2/s"),
            "CHI_E_ANOM": ("SIGNAL", "anomalous electron heat conductivity, m^2/s"),
            "CHI_I_ANOM": ("SIGNAL", "anomalous ion heat conductivity, m^2/s"),
            "CHI_E_NEO": ("SIGNAL", "neoclassical electron heat conductivity, m^2/s"),
            "CHI_I_NEO": ("SIGNAL", "neoclassical ion heat conductivity, m^2/s"),
            "DIFF": ("SIGNAL", "diffusion coefficient, m^2/s"),
            "QE": ("SIGNAL", "electron power flux, MW"),
            "QI": ("SIGNAL", "ion power flux, MW"),
            "QN": ("SIGNAL", "total electron flux, 10^19/s"),
            "PEGN": ("SIGNAL", "electron convective heat flux, MW"),
            "PIGN": ("SIGNAL", "ion convective heat flux, MW"),
            "CC": ("SIGNAL", "Parallel current conductivity, 1/(Ohm*m),"),
            "ELON": ("SIGNAL", "Elongation profile"),
            "TRI": ("SIGNAL", "Triangularity (up/down symmetrized), profile"),
            "RMID": ("SIGNAL", "Centre of flux surfaces, m"),
            "RMINOR": ("SIGNAL", "minor radius, m"),
            "N_D": ("SIGNAL", "Deuterium density,10E19/m3"),
            "N_T": ("SIGNAL", "Tritium density	,10E19/m3"),
            "T_D": ("SIGNAL", "Deuterium temperature,keV"),
            "T_T": ("SIGNAL", "Tritium temperature,keV"),
            "Q_RF": ("SIGNAL", "RF power density to electron,MW/m3"),
            "Q_ALPHA_E": ("SIGNAL", "Alpha power density to electrons,MW/m3"),
            "Q_ALPHA_I": ("SIGNAL", "Alpha power density to ions,MW/m3"),
            "J_NBI": ("SIGNAL", "NB driven current density,MA/m2"),
            "J_RF": ("SIGNAL", " EC driven current density,MA/m2"),
            "J_BS": ("SIGNAL", "Bootstrap current density,MA/m2"),
            "J_OH": ("SIGNAL", "Ohmic current density,MA/m2"),
            "J_TOT": ("SIGNAL", "Total current density,MA/m2"),
            "PSIN": ("SIGNAL", "Normalized poloidal flux -"),
            "CN": ("SIGNAL", "Particle pinch velocity , m/s"),
            "SBM": ("SIGNAL", "Particle source from beam, 10^19/m^3/s "),
            "SPEL": ("SIGNAL", "Particle source from pellets, 10^19/m^3/s"),
            "SWALL": ("SIGNAL", "Particle source from wall neutrals, 10^19/m^3/s"),
            "OMEGA_TOR": ("SIGNAL", "Toroidal rotation frequency, 1/s"),
            "TORQ_DEN": ("SIGNAL", "Total torque density from NB, N*m/m3"),
            "TORQ_DEN_BE": (
                "SIGNAL",
                "Collisional to electron torque density from NB, N*m/m3",
            ),
            "TORQ_DEN_BI": (
                "SIGNAL",
                "Collisional to ions torque density from NB, N*m/m3",
            ),
            "TORQ_DEN_BTH": (
                "SIGNAL",
                "Beam thermalisation torque density from NB, N*m/m3",
            ),
            "TORQ_DEN_JXB": ("SIGNAL", "JXB torque density from NB, N*m/m3"),
            "TORQ_DEN_BCX": ("SIGNAL", "CX losses torque density from NB, N*m/m3"),
            "CHI_PHI": ("SIGNAL", "Momentum transport coefficient, m2/s"),
        },
        "P_BOUNDARY": {
            "INDEX": ("NUMERIC", ""),
            "RBND": ("SIGNAL", "R OF PLASMA_BOUNDARY"),
            "ZBND": ("SIGNAL", "Z OF PLASMA_BOUNDARY"),
        },
        "P_BOUNDARY.XPOINTS": {
            "FXPM": ("SIGNAL", "x-point poloidal flux, Wb"),
            "RXPM": ("SIGNAL", "r-position,m"),
            "ZXPM": ("SIGNAL", "z-position,m"),
            "ACTIVE": ("SIGNAL", "=1 if divertor"),
        },
        "P_BOUNDARY.LIMITER.VESSEL": {
            "INDEX": ("NUMERIC", ""),
            "R": ("NUMERIC", "r-position,m"),
            "Z": ("NUMERIC", "z-position,m"),
        },
        "P_BOUNDARY.LIMITER.PLASMA": {
            "ACTIVE": ("SIGNAL", "=1 if limiter"),
            "FBND": ("SIGNAL", "Limiter poloidal flux, Wb"),
            "R": ("SIGNAL", "r-position,m"),
            "Z": ("SIGNAL", "z-position,m"),
        },
        "P_BOUNDARY.TARGETS": {
            "RTORX": ("SIGNAL", "Geometric axis r-position from ASTRA exp file,m"),
            "ZX": ("SIGNAL", "Geometric axis z-position from ASTRA exp file,m"),
            "ELONGX": ("SIGNAL", "Elongation from ASTRA exp file"),
            "TRIANX": ("SIGNAL", "Triangularity from ASTRA exp file"),
            "ABCX": ("SIGNAL", "Minor radius at midplane from ASTRA exp file,m"),
        },
    }

    return nodes


def hda():

    # TODO: create upper and lower bounds for global quantities

    nodes = {
        "": {
            "TIME": ("NUMERIC", "time vector, s"),
        },  # (type, description)
        ".METADATA": {
            "USER": ("TEXT", "Username of owner"),
            "PULSE": ("NUMERIC", "Pulse number analysed"),
            "EQUIL": ("TEXT", "Equilibrium used"),
            "EL_DENS": ("TEXT", "Electron density diagnostic used for optimization"),
            "EL_TEMP": (
                "TEXT",
                "Electron temperature diagnostic used for optimization",
            ),
            "ION_TEMP": ("TEXT", "Ion temperature diagnostic used for optimization"),
            "STORED_EN": ("TEXT", "Stored energy diagnostic used for optimization"),
            "MAIN_ION": ("TEXT", "Main ion element"),
            "IMPURITY1": ("TEXT", "Impurity element chosen for Z1"),
            "IMPURITY2": ("TEXT", "Impurity element chosen for Z2"),
            "IMPURITY3": ("TEXT", "Impurity element chosen for Z3"),
        },
        ".GLOBAL": {
            "CR0": ("SIGNAL", "Minor radius = (R_LFS - R_HFS)/2 at midplane, m"),
            "RMAG": ("SIGNAL", "Magnetic axis R, m"),
            "ZMAG": ("SIGNAL", "Magnetic axis z, m"),
            "VOLM": ("SIGNAL", "Plasma volume z, m^3"),
            "IP": ("SIGNAL", "Plasma current, A"),
            "TE0": ("SIGNAL", "Central electron temp, eV"),
            "TI0": ("SIGNAL", "Central main ion temp, eV"),
            "TI0_Z1": ("SIGNAL", "Central impurity1 ion temp, eV"),
            "TI0_Z2": ("SIGNAL", "Central impurity2 ion temp, eV"),
            "TI0_Z3": ("SIGNAL", "Central impurity3 ion temp, eV"),
            "NE0": ("SIGNAL", "Central electron density, m^-3 "),
            "NI0": ("SIGNAL", "Central main ion density, m^-3 "),
            "TEV": ("SIGNAL", "Volume average electron temp, eV"),
            "TIV": ("SIGNAL", "Volume average ion temp, eV"),
            "NEV": ("SIGNAL", "Volume average electron density m^-3"),
            "NIV": ("SIGNAL", "Volume average main ion density m^-3"),
            "WP": ("SIGNAL", "Total stored energy, J"),
            "WTH": ("SIGNAL", "Thermal stored energy, J"),
            "UPL": ("SIGNAL", "Loop Voltage, V"),
            "P_OH": ("SIGNAL", "Total Ohmic power, W"),
            "ZEFF": ("SIGNAL", "Effective charge at the plasma center"),
            "ZEFFV": ("SIGNAL", "Volume averaged effective charge"),
            "CION": ("SIGNAL", "Average concentration of main ion"),
            "CIM1": ("SIGNAL", "Average concentration of impurity IMP1"),
            "CIM2": ("SIGNAL", "Average concentration of impurity IMP2"),
            "CIM3": ("SIGNAL", "Average concentration of impurity IMP3"),
        },
        ".PROFILES.PSI_NORM": {
            "RHOP": ("NUMERIC", "Radial vector, Sqrt of normalised poloidal flux"),
            "XPSN": ("NUMERIC", "x vector - fi_normalized"),
            "P": ("SIGNAL", "Pressure,Pa"),
            "P_HI": ("SIGNAL", "Pressure upper bound,Pa"),
            "P_LO": ("SIGNAL", "Pressure lower bound,Pa"),
            "VOLUME": ("SIGNAL", "Volume inside magnetic surface,m^3"),
            "NE": ("SIGNAL", "Electron density, m^-3"),
            "NE_HI": ("SIGNAL", "Electron density upper limit, m^-3"),
            "NE_LO": ("SIGNAL", "Electron density lower limit, m^-3"),
            "NI": ("SIGNAL", "Ion density, m^-3"),
            "NI_HI": ("SIGNAL", "Ion density upper limit, m^-3"),
            "NI_LO": ("SIGNAL", "Ion density lower limit, m^-3"),
            "TE": ("SIGNAL", "Electron temperature, eV"),
            "TE_HI": ("SIGNAL", "Electron temperature upper limit, eV"),
            "TE_LO": ("SIGNAL", "Electron temperature lower limit, eV"),
            "TI": ("SIGNAL", "Ion temperature of main ion, eV"),
            "TI_HI": ("SIGNAL", "Ion temperature of main ion upper limit, eV"),
            "TI_LO": ("SIGNAL", "Ion temperature of main ion lower limit, eV"),
            "TIZ1": ("SIGNAL", "Ion temperature of impurity IMP1, eV"),
            "TIZ1_HI": ("SIGNAL", "Ion temperature of impurity IMP1 upper limit, eV"),
            "TIZ1_LO": ("SIGNAL", "Ion temperature of impurity IMP1 lower limit, eV"),
            "TIZ2": ("SIGNAL", "Ion temperature of impurity IMP2, eV"),
            "TIZ2_HI": ("SIGNAL", "Ion temperature of impurity IMP2 upper limit, eV"),
            "TIZ2_LO": ("SIGNAL", "Ion temperature of impurity IMP2 lower limit, eV"),
            "TIZ3": ("SIGNAL", "Ion temperature of impurity IMP3, eV"),
            "TIZ3_HI": ("SIGNAL", "Ion temperature of impurity IMP3 upper limit, eV"),
            "TIZ3_LO": ("SIGNAL", "Ion temperature of impurity IMP3 lower limit, eV"),
            "NIZ1": ("SIGNAL", "Density of impurity IMP1, m^-3"),
            "NIZ1_HI": ("SIGNAL", "Density of impurity IMP1 upper limit, m^-3"),
            "NIZ1_LO": ("SIGNAL", "Density of impurity IMP1 lower limit, m^-3"),
            "NIZ2": ("SIGNAL", "Density of impurity IMP2, m^-3"),
            "NIZ2_HI": ("SIGNAL", "Density of impurity IMP2 upper limit, m^-3"),
            "NIZ2_LO": ("SIGNAL", "Density of impurity IMP2 lower limit, m^-3"),
            "NIZ3": ("SIGNAL", "Density of impurity IMP3, m^-3"),
            "NIZ3_HI": ("SIGNAL", "Density of impurity IMP3 upper limit, m^-3"),
            "NIZ3_LO": ("SIGNAL", "Density of impurity IMP3 lower limit, m^-3"),
            "NNEUTR": ("SIGNAL", "Density of neutral main ion, m^-3"),
            "NNEUTR_HI": ("SIGNAL", "Density of neutral main ion upper limit, m^-3"),
            "NNEUTR_LO": ("SIGNAL", "Density of neutral main ion lower limit, m^-3"),
            "ZI": ("SIGNAL", "Average charge of main ion, "),
            "ZI_HI": ("SIGNAL", "Average charge of main ion upper bound, "),
            "ZI_LO": ("SIGNAL", "Average charge of main ion lower bound, "),
            "ZIM1": ("SIGNAL", "Average charge of impurity IMP1, "),
            "ZIM1_HI": ("SIGNAL", "Average charge of impurity IMP1 upper bound, "),
            "ZIM1_LO": ("SIGNAL", "Average charge of impurity IMP1 lower bound, "),
            "ZIM2": ("SIGNAL", "Average charge of impurity IMP2, "),
            "ZIM2_HI": ("SIGNAL", "Average charge of impurity IMP2 upper bound, "),
            "ZIM2_LO": ("SIGNAL", "Average charge of impurity IMP2 lower bound, "),
            "ZIM3": ("SIGNAL", "Average charge of impurity IMP3, "),
            "ZIM3_HI": ("SIGNAL", "Average charge of impurity IMP3 upper bound, "),
            "ZIM3_LO": ("SIGNAL", "Average charge of impurity IMP3 lower bound, "),
            "ZEFF": ("SIGNAL", "Effective charge, "),
            "ZEFF_HI": ("SIGNAL", "Effective charge upper limit, "),
            "ZEFF_LO": ("SIGNAL", "Effective charge lower limit, "),
        },
        ".PROFILES.R_MIDPLANE": {
            "RPOS": ("NUMERIC", "Major radius position of measurement"),
            "ZPOS": ("NUMERIC", "Z position of measurement"),
            "P": ("SIGNAL", "Pressure,Pa"),
            "VOLUME": ("SIGNAL", "Volume inside magnetic surface,m^3"),
            "NE": ("SIGNAL", "Electron density, m^-3"),
            "NI": ("SIGNAL", "Ion density, m^-3"),
            "TE": ("SIGNAL", "Electron temperature, eV"),
            "TI": ("SIGNAL", "Ion temperature of main ion, eV"),
            "TIZ1": ("SIGNAL", "Ion temperature of impurity IMP1, eV"),
            "TIZ2": ("SIGNAL", "Ion temperature of impurity IMP2, eV"),
            "TIZ3": ("SIGNAL", "Ion temperature of impurity IMP3, eV"),
            "NIZ1": ("SIGNAL", "Density of impurity IMP1, m^-3"),
            "NIZ2": ("SIGNAL", "Density of impurity IMP2, m^-3"),
            "NIZ3": ("SIGNAL", "Density of impurity IMP3, m^-3"),
            "NNEUTR": ("SIGNAL", "Density of neutral main ion, m^-3"),
            "ZI": ("SIGNAL", "Average charge of main ion, "),
            "ZIM1": ("SIGNAL", "Average charge of impurity IMP1, "),
            "ZIM2": ("SIGNAL", "Average charge of impurity IMP2, "),
            "ZIM3": ("SIGNAL", "Average charge of impurity IMP3, "),
            "ZEFF": ("SIGNAL", "Effective charge, "),
        },
    }
    # "RHOT": ("SIGNAL", "Sqrt of normalised toroidal flux, xpsn"),

    return nodes



# TODO: implement test to check availability of instruments methods and
#       quantities have corresponding DATATYPE
# TODO: why a class and not a simple JSON file or a dictionary?

from abc import ABC
from typing import Dict
from typing import Tuple


class MachineConf(ABC):
    """Machine configuration abstract class containing:
    MACHINE_DIMS = ((R_min, R_ma), (z_min, z_max))
    INSTRUMENT_METHODS = {"instrument_name": "reader_method"}
    QUANTITIES_PATH = {"reader_method": {"quantity":"database_path"}
    """

    def __init__(self):
        self.MACHINE_DIMS: Tuple[Tuple[float, float], Tuple[float, float]] = ()
        self.INSTRUMENT_METHODS: Dict[str, str] = {}
        self.QUANTITIES_PATH: Dict[str, Dict[str, str]] = {}



import os.path

import matplotlib.pylab as plt
import numpy as np
from scipy.interpolate import CubicSpline
import yaml

import indica
from indica.profilers.profiler_base import ProfilerBase
from indica.utilities import format_coord
from indica.utilities import format_dataarray


def gaussian(x, A, B, x_0, w):
    return (A - B) * np.exp(-((x - x_0) ** 2) / (2 * w**2)) + B


class ProfilerGauss(ProfilerBase):
    def __init__(
        self,
        datatype: str = "electron_temperature",
        xend: float = 1.05,
        xspl: np.ndarray = None,
        coord="rhop",
        parameters: dict = None,
    ):
        """
        Class to build general gaussian profiles

        Parameters
        ----------
        datatype
            str defining what type of profile is to be built
        xspl
            normalised radial grid [0, 1] on which profile is to be built

        """
        super().__init__(parameters)
        self.y0: float = None
        self.y1: float = None
        self.yend: float = None
        self.peaking: float = None
        self.wcenter: float = None
        self.wped: float = None
        self.xend = xend
        self.coord = coord
        self.x = np.linspace(0, 1, 15) ** 0.7
        self.datatype = datatype
        if xspl is None:
            xspl = format_coord(np.linspace(0, 1.0, 30), self.coord)
        self.xspl = xspl

        default_parameters = get_defaults_for_profiler_gauss(datatype=datatype)
        parameters = dict(default_parameters, **self.parameters)  # Overwrites defaults
        self.set_parameters(**parameters)

    def __call__(
        self,
        y0_fix=False,
        y0_ref=None,
        wcenter_exp=0.05,
        debug=False,
    ):
        """
        Builds the profiles using the parameters set

        Parameters
        ----------
        y0_fix
            re-scale new profile to have central value = y0
        y0_ref
            reference y0 value
        wcenter_exp
            exponent for the additional peaking if y0_ref is not None

        Returns
        -------

        """

        # Add additional peaking with respect to reference shape
        peaking2 = 1.0
        if y0_ref is not None:
            if y0_ref < self.y0:
                peaking2 = self.y0 / y0_ref
        self.peaking2 = peaking2

        center = self.y0
        edge = self.y1

        if peaking2 > 1:
            center = y0_ref
            wcenter = self.wcenter - (peaking2**wcenter_exp - 1)
        else:
            wcenter = self.wcenter

        center = center / self.peaking

        x = self.x[np.where(self.x <= 1.0)[0]]

        # baseline profile shape
        y_baseline = (center - self.y1) * (1 - x**self.wped) + self.y1

        if self.peaking != 1:  # add central peaking
            sigma = wcenter / (np.sqrt(2 * np.log(2)))
            y_peaking1 = (
                gaussian(x, center * (self.peaking - 1), 0, 0, sigma) + y_baseline
            )
        else:
            y_peaking1 = y_baseline

        if peaking2 != 1:  # add additional peaking
            sigma = wcenter / (np.sqrt(2 * np.log(2)))
            y_peaking2 = (
                gaussian(x, y_peaking1[0] * (peaking2 - 1), 0, 0, sigma) + y_peaking1
            )
        else:
            y_peaking2 = y_peaking1

        if y0_fix:
            y = y_peaking2 - edge
            y = y / y[0]
            y = y * (center - edge)
            y = y + edge

        else:
            y = y_peaking2

        x = np.append(x, self.xend)
        y = np.append(y, self.yend)

        self.cubicspline = CubicSpline(
            x,
            y,
            0,
            "clamped",
            False,
        )
        _yspl = self.cubicspline(self.xspl)
        coords = {self.coord: self.xspl}
        ydata = format_dataarray(_yspl, self.datatype, coords=coords)
        self.ydata = ydata

        if debug:
            plt.figure()
            plt.plot(x, y_baseline, label="first")
            plt.plot(x, y_peaking1, label="peaking1")
            plt.plot(x, y_peaking2, label="peaking2")
            plt.plot(x, y, label="y0_fix", marker="o")
            plt.plot(self.xspl, ydata, label="spline")
            plt.legend()

        return ydata


def get_defaults_for_profiler_gauss(
    datatype="electron_temperature", config_name="profiler_gauss"
):
    """
    Loads config for default parameter values
    """
    path = os.path.join(
        os.path.dirname(indica.__file__), f"configs/profilers/{config_name}.yaml"
    )
    with open(path) as stream:
        cfg = yaml.safe_load(stream)
    return cfg[datatype]


def initialise_gauss_profilers(
    x_coord, profile_names: list, profile_params: dict = None
):
    if profile_params is None:
        profile_params = {}
    _profilers = {
        profile_name: ProfilerGauss(
            datatype=profile_name.split(":")[0],
            xspl=x_coord,
            parameters=profile_params.get(profile_name, {}),
        )
        for profile_name in profile_names
    }
    return _profilers



"""Classes to convert between coordinate systems. Also contains
routines for interpolating or downsampling in time."""

from .abstractconverter import CoordinateTransform
from .line_of_sight import LineOfSightTransform
from .spot_weightings import SpotWeightings
from .time import bin_to_time_labels
from .time import convert_in_time
from .transect import TransectCoordinates
from .trivial import TrivialTransform

__all__ = [
    "CoordinateTransform",
    "LineOfSightTransform",
    "SpotWeightings",
    "TransectCoordinates",
    "TrivialTransform",
    "bin_to_time_labels",
    "convert_in_time",
]
